#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for utils.py
"""

import PanACoTA.utils as utils
import test.test_unit.utilities_for_tests as utilities
import pytest
import os
import logging
import shutil
import matplotlib

matplotlib.use('AGG')

# Define variables used by several tests
DATA_DIR = os.path.join("test", "data", "annotate")
TEST_DIR = os.path.join(DATA_DIR, "test_files")
BASELINE_DIR = os.path.abspath(os.path.join(DATA_DIR, "exp_files", "baseline"))
GENEPATH = os.path.join(DATA_DIR, "generated_by_unit-tests")
LOGFILE_BASE = "test_prokka"
LOGFILES = [LOGFILE_BASE + ext for ext in [".log", ".log.debug", ".log.details", ".log.err"]]

@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module
    """
    # Init logger to level detail (15)
    # utils.init_logger(LOGFILE_BASE, logging.DEBUG, 'test_utils', verbose=1)
    os.mkdir(GENEPATH)
    print("setup")

    yield
    shutil.rmtree(GENEPATH)
    print("teardown")


# Start tests
def test_check_install():
    """
    Try to run prokka, which is installed, and check that there is no problem
    """
    assert utils.check_installed("prokka")


def test_check_install_error():
    """
    Try to run a command which does not exist, and check that it closes the program
    with exit code 1
    """
    assert not utils.check_installed("plop false command...")


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_DIR, tolerance=5, backend="agg")
def test_plot_dist():
    """
    Plot a given distribution, and check that output is as expected
    """
    logger = logging.getLogger("test_utils")
    values = [1, 1, 1, 1, 1, 1, 2, 2, 2, 3, 3, 10]
    limit = 3
    res_dir = os.path.join("test", "data", "annotate")
    # reffile = os.path.join("test", "data", "annotate", "exp_files", "res_plot_distr.png")
    title = "Distribution test"
    text = "Max L90 ="
    myfig = utils.plot_distr(values, limit, title, text, logger)
    return myfig


def test_skipped_prokka_prodigal(caplog):
    """
    Test that when the list of skipped genomes (because of prokka run) is not empty,
    it writes the right message.
    """
    caplog.set_level(logging.DEBUG)
    skipped = ["toto", "genome", "genome2"]
    utils.write_warning_skipped(skipped)
    assert ("prokka had problems while annotating some genomes, or did not "
            "find any gene. Hence, they are not "
            "formatted, and absent from your output database. Please look at the "
            "current error log (<output_directory>/PanACoTA-annotate_list_genomes[-date].log.err) "
            "to get more information on the problems. "
            "Here are those genomes:") in caplog.text
    assert "\n\t- toto\n\t- genome\n\t- genome2" in caplog.text
    utils.write_warning_skipped(skipped, prodigal_only=True)
    assert ("prodigal had problems while annotating some genomes, or did not "
            "find any gene. Hence, they are not "
            "formatted, and absent from your output database. Please look at the "
            "current error log (<output_directory>/PanACoTA-annotate_list_genomes[-date].log.err) "
            "to get more information on the problems. "
            "Here are those genomes:") in caplog.text
    assert "\n\t- toto\n\t- genome\n\t- genome2" in caplog.text


def test_skipped_format(caplog):
    """
    Test that when the list of skipped genomes (format step could not run) is not empty,
    it writes the right message.
    """
    caplog.set_level(logging.DEBUG)
    skipped_format = ["toto", "genome", "genome2"]
    utils.write_warning_skipped(skipped_format, do_format=True)
    assert ("Some genomes were annotated by prokka, but could not be formatted, "
            "and are hence absent from your output database. Please look at "
            "'<output_directory>/PanACoTA-annotate_list_genomes[-date].log.err' and .details "
            "files to get more information about why they could not be formatted") in caplog.text
    assert ("\n\t- toto\n\t- genome\n\t- genome2\n") in caplog.text


def test_write_discarded(caplog):
    """
    Test that the list of discarded genomes is written as expected.
    """
    caplog.set_level(logging.DEBUG)
    gnames = ["H299_H561.fasta", "B2_A3_5.fasta-problems", "genome1", "genome2", "genome3"]
    gpaths = [os.path.join(DATA_DIR, "genomes", name) for name in gnames]
    genomes = {gnames[0]: ["toto.0417", gpaths[0], gpaths[0], 12656, 3, 1],
               gnames[1]: ["toto.0417", gpaths[1], gpaths[1], 456464645, 5, 1],
               gnames[2]: ["toto.0417", gpaths[2], gpaths[2], 4564855, 156, 40],
               gnames[3]: ["toto.0417", gpaths[3], gpaths[3], 6549, 16, 8],
               gnames[4]: ["toto.0417", gpaths[4], gpaths[4], 9876546, 6, 2]
               }
    kept_genomes = ["H299_H561.fasta", "B2_A3_5.fasta-problems", "genome3"]
    list_file = os.path.join("titi", "toto", "list_genomes.txt")
    utils.write_genomes_info(genomes, kept_genomes, list_file, GENEPATH, qc=False)
    outfile = os.path.join(GENEPATH, "discarded-list_genomes.lst")
    exp_file = os.path.join(DATA_DIR, "exp_files", "res_test_write_discard.lst")
    assert "2 genomes were discarded" in caplog.text
    # There is no order in the discarded file. So, just check that the lines
    # written are as expected.
    assert utilities.compare_file_content(outfile, exp_file)


def test_write_discarded_1genome(caplog):
    """
    Test that the list of discarded genomes is written as expected.
    """
    caplog.set_level(logging.DEBUG)
    gnames = ["H299_H561.fasta", "B2_A3_5.fasta-problems", "genome1", "genome2", "genome3"]
    gpaths = [os.path.join(DATA_DIR, "genomes", name) for name in gnames]
    genomes = {gnames[0]: ["toto.0417", gpaths[0], gpaths[0], 12656, 3, 1],
               gnames[1]: ["toto.0417", gpaths[1], gpaths[1], 456464645, 5, 1],
               gnames[2]: ["toto.0417", gpaths[2], gpaths[2], 4564855, 156, 40],
               gnames[3]: ["toto.0417", gpaths[3], gpaths[3], 6549, 16, 8],
               gnames[4]: ["toto.0417", gpaths[4], gpaths[4], 9876546, 6, 2]
               }
    kept_genomes = ["H299_H561.fasta", "B2_A3_5.fasta-problems", "genome3", "genome2"]
    list_file = os.path.join("titi", "toto", "list_genomes.txt")
    utils.write_genomes_info(genomes, kept_genomes, list_file, GENEPATH, qc=False)
    outfile = os.path.join(GENEPATH, "discarded-list_genomes.lst")
    exp_file = os.path.join(DATA_DIR, "exp_files", "res_test_write_discard_1genome.lst")
    assert "1 genome was discarded" in caplog.text
    # There is no order in the discarded file. So, just check that the lines
    # written are as expected.
    assert utilities.compare_file_content(outfile, exp_file)


def test_write_discarded_qc(caplog):
    """
    Test that the list with information on all genomes when we run with QC only is
    written as expected
    """
    caplog.set_level(logging.DEBUG)
    gnames = ["H299_H561.fasta", "B2_A3_5.fasta-problems", "genome1", "genome2", "genome3"]
    gpaths = [os.path.join(DATA_DIR, "genomes", name) for name in gnames]
    genomes = {gnames[0]: ["toto.0417", gpaths[0], gpaths[0], 12656, 3, 1],
               gnames[1]: ["toto.0417", gpaths[1], gpaths[1], 456464645, 5, 1],
               gnames[2]: ["toto.0417", gpaths[2], gpaths[2], 4564855, 156, 40],
               gnames[3]: ["toto.0417", gpaths[3], gpaths[3], 6549, 16, 8],
               gnames[4]: ["toto.0417", gpaths[4], gpaths[4], 9876546, 6, 2]
               }
    kept_genomes = []
    list_file = os.path.join(DATA_DIR, "list_genomes.txt")
    utils.write_genomes_info(genomes, kept_genomes, list_file, GENEPATH, qc=True)
    outfile = os.path.join(GENEPATH, "ALL-GENOMES-info-list_genomes.lst")
    exp_file = os.path.join(DATA_DIR, "exp_files", "res_test_write_info_qc.lst")
    assert ("Writing information on genomes in "
            "test/data/annotate/generated_by_unit-tests/ALL-GENOMES-info-list_genomes.lst") in caplog.text
    # There is no order in the discarded file. So, just check that the lines
    # written are as expected.
    assert utilities.compare_file_content(outfile, exp_file)


def test_write_discarded_empty():
    """
    Test that when the list of genomes is empty, but the list of kept-genomes is not
    (should never happen...), it writes only the header of discarded lst file.
    """
    genomes = {}
    kept_genomes = ["H299_H561.fasta", "B2_A3_5.fasta-problems", "genome3"]
    list_file = os.path.join(DATA_DIR, "list_genomes.txt")
    utils.write_genomes_info(genomes, kept_genomes, list_file, GENEPATH)
    outfile = os.path.join(GENEPATH, "discarded-list_genomes.lst")
    with open(outfile, "r") as outf:
        all_lines = outf.readlines()
        assert len(all_lines) == 1
        assert all_lines[0] == "orig_name\tto_annotate\tgsize\tnb_conts\tL90\n"


def test_write_discarded_all_kept(caplog):
    """
    Test that when all genomes are kept, the discarded lst file only contains the
    header line.
    """
    caplog.set_level(logging.DEBUG)
    gnames = ["H299_H561.fasta", "B2_A3_5.fasta-problems", "genome1", "genome2", "genome3"]
    gpaths = [os.path.join(DATA_DIR, "genomes", name) for name in gnames]
    genomes = {gnames[0]: ["toto.0417", gpaths[0], gpaths[0], 12656, 3, 1],
               gnames[1]: ["toto.0417", gpaths[1], gpaths[1], 456464645, 5, 1],
               gnames[2]: ["toto.0417", gpaths[2], gpaths[2], 4564855, 156, 40],
               gnames[3]: ["toto.0417", gpaths[3], gpaths[3], 6549, 16, 8],
               gnames[4]: ["toto.0417", gpaths[4], gpaths[4], 9876546, 6, 2]
               }
    kept_genomes = ["H299_H561.fasta", "B2_A3_5.fasta-problems", "genome3", "genome2", "genome1"]
    list_file = os.path.join(DATA_DIR, "list_genomes.txt")
    utils.write_genomes_info(genomes, kept_genomes, list_file, GENEPATH)
    outfile = os.path.join(GENEPATH, "discarded-list_genomes.lst")
    with open(outfile, "r") as outf:
        all_lines = outf.readlines()
        assert len(all_lines) == 1
        assert all_lines[0] == "orig_name\tto_annotate\tgsize\tnb_conts\tL90\n"
    assert "0 genome was discarded" in caplog.text


def test_write_lstinfo():
    """
    Test that lstinfo file is written as expected.
    """
    gnames = ["H299_H561.fasta", "B2_A3_5.fasta-problems", "genome1", "genome2", "genome3"]
    gpaths = [os.path.join(DATA_DIR, "genomes", name) for name in gnames]
    genomes = {gnames[0]: ["toto.0417", gpaths[0], gpaths[0], 12656, 3, 1],
               gnames[1]: ["toto.0417", gpaths[1], gpaths[1], 456464645, 5, 1],
               gnames[2]: ["toto.0417", gpaths[2], gpaths[2], 4564855, 156, 40],
               gnames[3]: ["toto.0417", gpaths[3], gpaths[3], 6549, 16, 8],
               gnames[4]: ["toto.0417", gpaths[4], gpaths[4], 9876546, 6, 2]
               }
    list_file = os.path.join("toto", "list_genomes.txt")
    utils.write_lstinfo(list_file, genomes, GENEPATH)
    outfile = os.path.join(GENEPATH, "LSTINFO-list_genomes.lst")
    exp_file = os.path.join(DATA_DIR, "exp_files", "res_test_write_lstinfo.lst")
    assert utilities.compare_order_content(outfile, exp_file)


def test_write_lstinfo_nogenome():
    """
    Test that when there is no genome fully annotated, lstinfo contains
    only header.
    """
    genomes = {}
    list_file = os.path.join("toto", "list_genomes.txt")
    utils.write_lstinfo(list_file, genomes, GENEPATH)
    outfile = os.path.join(GENEPATH, "LSTINFO-list_genomes.lst")
    with open(outfile, "r") as outf:
        all_lines = outf.readlines()
        assert len(all_lines) == 1
        assert all_lines[0] == "gembase_name\torig_name\tto_annotate\tgsize\tnb_conts\tL90\n"


def test_sort_gene():
    """
    Test that genomes are sorted by species first, and then by strain number.
    """
    # genomes = {genome_orig, [gembase, path, gsize, nbcont, L90]}
    genomes = ["genome.0417.00010", "toto.0417.00010", "genome1.0417.00002",
               "genome.0417.00015", "totn.0417.00010", "genome.0417.00009",
               "genome.0517.00001", "toto.0417.00011"]
    sorted_genomes = sorted(genomes, key=utils.sort_genomes_by_name)
    exp = ["genome.0517.00001", "genome.0417.00009", "genome.0417.00010",
           "genome.0417.00015", "genome1.0417.00002", "totn.0417.00010",
           "toto.0417.00010", "toto.0417.00011"]
    assert sorted_genomes == exp


def test_sort_gene_tuple():
    """
    Test that genomes are sorted by species first, and then by strain number.
    """
    # genomes = {genome_orig, [gembase, path, gsize, nbcont, L90]}
    genomes = {"name1": ["genome.0417.00010", "path/to/genome", 123456, 50, 3],
               "name2": ["toto.0417.00010", "path/to/genome", 123456, 50, 3],
               "name3": ["genome1.0417.00002", "path/to/genome", 123456, 50, 3],
               "name4": ["genome.0417.00015", "path/to/genome", 123456, 50, 3],
               "name5": ["totn.0417.00010", "path/to/genome", 123456, 50, 3],
               "name6": ["genome.0417.00009", "path/to/genome", 123456, 50, 3],
               "name7": ["genome.0517.00001", "path/to/genome", 123456, 50, 3],
               "name8": ["toto.0417.00011", "path/to/genome", 123456, 50, 3], }
    sorted_genomes = sorted(genomes.items(), key=utils.sort_genomes_by_name)
    exp = [("name7", genomes["name7"]), ("name6", genomes["name6"]),
           ("name1", genomes["name1"]), ("name4", genomes["name4"]),
           ("name3", genomes["name3"]), ("name5", genomes["name5"]),
           ("name2", genomes["name2"]), ("name8", genomes["name8"])]
    assert sorted_genomes == exp


def test_sort_gene_noformat():
    """
    Test that when genomes are not in the gembase format, it returns
    genomes in alphabetical order
    """
    # genomes = {genome_orig, [gembase, path, gsize, nbcont, L90]}
    genomes = {"name1": ["genome", "path/to/genome", 123456, 50, 3],
               "name2": ["toto", "path/to/genome", 123456, 50, 3],
               "name3": ["genome1   ", "path/to/genome", 123456, 50, 3],
               "name4": ["genome1bis", "path/to/genome", 123456, 50, 3],
               "name5": ["mygenome", "path/to/genome", 123456, 50, 3],
               "name6": ["agenome_nogembase", "path/to/genome", 123456, 50, 3],
               "name7": ["agenome_nogembase2", "path/to/genome", 123456, 50, 3],
               "name8": ["toto.0417.00011", "path/to/genome", 123456, 50, 3], }
    sorted_genomes = sorted(genomes.items(), key=utils.sort_genomes_by_name)
    exp = [("name6", genomes["name6"]), ("name7", genomes["name7"]),
           ("name1", genomes["name1"]), ("name3", genomes["name3"]),
           ("name4", genomes["name4"]), ("name5", genomes["name5"]),
           ("name2", genomes["name2"]), ("name8", genomes["name8"])]
    assert sorted_genomes == exp


def test_sort_genome_l90():
    """
    Test that when genomes are not in the gembase format, it returns
    genomes in alphabetical order
    """
    # genomes = {genome_orig, [gembase, path, gsize, nbcont, L90]}
    genomes = {"name1": ["genome", "path/to/genome", 123456, 1, 2],
               "name2": ["toto", "path/to/genome", 123456, 100, 4],
               "name3": ["genome1   ", "path/to/genome", 123456, 10, 2],
               "name4": ["genome1bis", "path/to/genome", 123456, 11, 2],
               "name5": ["mygenome", "path/to/genome", 123456, 8, 3],
               "name6": ["agenome_nogembase", "path/to/genome", 123456, 3, 1],
               "name7": ["agenome_nogembase2", "path/to/genome", 123456, 4, 1],
               "name8": ["toto.0417.00011", "path/to/genome", 123456, 50, 5], }
    sorted_genomes = sorted(genomes.items(), key=utils.sort_genomes_l90_nbcont)
    exp = [("name6", genomes["name6"]), ("name7", genomes["name7"]),
           ("name1", genomes["name1"]), ("name3", genomes["name3"]),
           ("name4", genomes["name4"]), ("name5", genomes["name5"]),
           ("name2", genomes["name2"]), ("name8", genomes["name8"])]
    assert sorted_genomes == exp


def test_sort_proteins():
    """
    Test that when a list of proteins is given, it returns the list sorted by :
    - species
    - strain number
    - protein number
    """
    proteins = ["ESCO.0417.00010.i0001_12354", "ESCA.0617.00001.i0001_005",
                "ESCO.0517.00001.i0001_12354", "ESCO.0417.00010.i0001_1",
                "TOTO.0417.00001.i0001_12", "TOTO.0717.00001.i0008_9"]
    sorted_prot = sorted(proteins, key=utils.sort_proteins)
    exp = ["ESCA.0617.00001.i0001_005", "ESCO.0517.00001.i0001_12354",
           "ESCO.0417.00010.i0001_1", "ESCO.0417.00010.i0001_12354",
           "TOTO.0717.00001.i0008_9", "TOTO.0417.00001.i0001_12"]
    assert sorted_prot == exp


def test_sort_proteins_other_format():
    """
    Test that when a list of proteins is given, it returns the list sorted by :
    - species
    - strain number
    - protein number
    for all proteins where this information is available,
    and then, the other proteins, sorted by :
    - species (string before "_")
    - protein number (num after "_")
    """
    proteins = ["ESCO.0417.00010.i0001_12354", "ESCO.0617.00001.i0001_005", "my_protein_0002",
                "ESCO.0517.00001.i0001_12354", "my_protein_0001", "ESCO.0417.00010.i0001_1",
                "a_my_prot_15", "a_my_prot_2",
                "TOTO.0417.00001.i0001_12", "TOTO.0717.00001.i0008_9"]
    sorted_prot = sorted(proteins, key=utils.sort_proteins)
    exp = ["ESCO.0617.00001.i0001_005", "ESCO.0517.00001.i0001_12354",
           "ESCO.0417.00010.i0001_1", "ESCO.0417.00010.i0001_12354",
           "TOTO.0717.00001.i0008_9", "TOTO.0417.00001.i0001_12",
           "a_my_prot_2", "a_my_prot_15", "my_protein_0001", "my_protein_0002"]
    assert sorted_prot == exp


def test_sort_proteins_error_format1(caplog):
    """
    Test that when a protein name does not follow the format <alpha_num>_<num>,
    it gives an error.
    """
    caplog.set_level(logging.DEBUG)
    proteins = ["ESCO.0417.00010.i0001_12354", "ESCO.0617.00001.i0001_005",
                "ESCO.0517.00001.i0001_12354", "error-protein"]
    with pytest.raises(SystemExit):
        sorted(proteins, key=utils.sort_proteins)
    assert ("ERROR: Protein error-protein does not have the required format. It must contain, "
            "at least <alpha-num>_<num_only>, and at best "
            "<name>.<date>.<strain_num>.<contig_info>_<prot_num>. "
            "Please change its name.") in caplog.text


def test_read_genomes_nofile(caplog):
    """
    Test that when the genome list file provided does not exist, it
    ends the program with an error message
    """
    logger = logging.getLogger("default")
    with pytest.raises(SystemExit):
        utils.read_genomes("toto.txt", "TOTO", "0417", "db/path", "tmppath", logger)
    assert "ERROR: Your list file " in caplog.text
    assert "toto.txt" in caplog.text
    assert "does not exist. Please provide a list file." in caplog.text
    assert "Ending program." in caplog.text


def test_read_genomes_wrongname():
    """
    Test that when the list file contains only genome names which do not exist,
    it returns an empty list of genomes to annotate/format.
    """
    logger = logging.getLogger("default")
    name = "ESCO"
    date = "0417"
    dbpath = os.path.join(DATA_DIR, "genomes")
    tmppath = "tmppath"
    list_file = os.path.join(DATA_DIR, "test_files", "list_genomes-wrongNames.txt")
    genomes = utils.read_genomes(list_file, name, date, dbpath, tmppath, logger)
    assert genomes == {}


def test_read_genomes_ok(caplog):
    """
    Test that when the list file contains genomes existing, it returns the expected list
    of genomes
    """
    logger = logging.getLogger("default")
    name = "ESCO"
    date = "0417"
    dbpath = os.path.join(DATA_DIR, "genomes")
    tmppath = "tmppath"
    list_file = os.path.join(DATA_DIR, "test_files", "list_genomes.lst")
    genomes = utils.read_genomes(list_file, name, date, dbpath, tmppath, logger)
    exp = {"A_H738.fasta": ["ESCO.0417"],
           "B2_A3_5.fasta-split5N.fna-short-contig.fna": ["ESCO.0417"],
           "H299_H561.fasta": ["ABCD.0417"], "genome2.fasta": ["TOTO.0417"],
           "genome3.fasta": ["ESCO.0512"], "genome4.fasta": ["TOTO.0417"],
           "genome5.fasta": ["TOTO.0114"]}
    assert exp == genomes
    assert "genome.fst genome file does not exist. It will be ignored." in caplog.text


def test_read_genomes_errors(caplog):
    """
    Test that when the list file contains errors in name and date provided,
    it returns the expected errors, and the expected genome list.
    """
    logger = logging.getLogger("default")
    name = "ESCO"
    date = "0417"
    dbpath = os.path.join(DATA_DIR, "genomes")
    tmppath = "tmppath"
    list_file = os.path.join(DATA_DIR, "test_files", "list_genomes-errors.txt")
    genomes = utils.read_genomes(list_file, name, date, dbpath, tmppath, logger)
    exp = {"A_H738.fasta": ["ESCO.0417"],
           "B2_A3_5.fasta-split5N.fna-short-contig.fna": ["ESCO.0417"],
           "H299_H561.fasta": ["ABCD.0417"], "genome2.fasta": ["TOTO.0417"],
           "genome3.fasta": ["ESCO.0512"], "genome4.fasta": ["ESCO.0417"],
           "genome5.fasta": ["ESCO.0417"]}
    assert genomes == exp
    assert ("Invalid name/date given for genome A_H738.fasta. Only put "
            "4 alphanumeric characters in your date and name. For "
            "this genome, the default name (ESCO) and date (0417) will "
            "be used.") in caplog.text
    assert ("Invalid name abc given for genome B2_A3_5.fasta-split5N.fna-short-contig.fna. Only put "
           "4 alphanumeric characters in your date and name. For "
           "this genome, the default name (ESCO) will "
           "be used.") in caplog.text
    assert ("Invalid date 152 given for genome H299_H561.fasta. Only put "
            "4 alphanumeric characters in your date and name. For "
            "this genome, the default date (0417) will "
            "be used.") in caplog.text
    assert ("Invalid date 1-03 given for genome genome2.fasta. Only put "
            "4 alphanumeric characters in your date and name. For "
            "this genome, the default date (0417) will "
            "be used.") in caplog.text
    assert ("genome.fst genome file does not exist. "
            "It will be ignored.") in caplog.text
    assert ("Invalid name a/b2 given for genome genome3.fasta. Only put "
            "4 alphanumeric characters in your date and name. For "
            "this genome, the default name (ESCO) will "
            "be used.") in caplog.text
    assert ("Invalid name #esc given for genome genome5.fasta. Only put "
            "4 alphanumeric characters in your date and name. For "
            "this genome, the default name (ESCO) will "
            "be used.") in caplog.text
    assert ("Invalid date 1_14 given for genome genome5.fasta. Only put "
            "4 alphanumeric characters in your date and name. For "
            "this genome, the default date (0417) will "
            "be used.") in caplog.text


def test_read_genomes_multi_files(caplog):
    """
    Test that when the list file contains several filenames for 1 same genome,
    it returns the expected genome list, the expected errors (when some genome
    files do not exist) and the expected concatenated files.
    """
    logger = logging.getLogger("default")
    name = "ESCO"
    date = "0417"
    tmppath = os.path.join(GENEPATH, "tmppath")
    os.mkdir(tmppath)
    dbpath = os.path.join(DATA_DIR, "genomes")
    list_file = os.path.join(DATA_DIR, "test_files", "list_genomes-multi-files.txt")
    genomes = utils.read_genomes(list_file, name, date, dbpath, tmppath, logger)
    exp = {"A_H738.fasta-all.fna": ["ESCO.0417"],
           "H299_H561.fasta-all.fna": ["ABCD.0417"], "genome2.fasta": ["TOTO.0417"],
           "genome3.fasta": ["ESCO.0512"], "genome4.fasta": ["TOTO.0417"],
           "genome5.fasta": ["TOTO.0114"]}
    assert exp == genomes
    # Check error messages
    assert ("genome.fna genome file does not exist. Its file will be ignored "
            "when concatenating ['H299_H561.fasta', 'genome6.fasta', 'genome.fna']") in caplog.text
    assert "genome.fst genome file does not exist. It will be ignored." in caplog.text
    assert ("toto.fst genome file does not exist. Its file will be ignored "
            "when concatenating ['toto.fst', 'toto.fasta', 'genome.fst']") in caplog.text
    assert ("toto.fasta genome file does not exist. Its file will be ignored "
            "when concatenating ['toto.fst', 'toto.fasta', 'genome.fst']") in caplog.text
    assert ("genome.fst genome file does not exist. Its file will be ignored "
            "when concatenating ['toto.fst', 'toto.fasta', 'genome.fst']") in caplog.text
    assert ("None of the genome files in ['toto.fst', 'toto.fasta', 'genome.fst'] exist. "
            "This genome will be ignored.") in caplog.text
    # Check that files were concatenated as expected
    concat1 = os.path.join(tmppath, "A_H738.fasta-all.fna")
    exp_concat1 = os.path.join(dbpath, "A_H738-and-B2_A3_5.fna")
    concat2 = os.path.join(tmppath, "H299_H561.fasta-all.fna")
    exp_concat2 = os.path.join(dbpath, "H299_H561-and-genome6.fna")
    assert utilities.compare_order_content(concat1, exp_concat1)
    assert utilities.compare_order_content(concat2, exp_concat2)


def test_read_genomes_info_nofile(caplog):
    """
    Read lstinfo file and get all genomes information
    Check that when the file does not exist, it exits with appropriate error message
    """
    caplog.set_level(logging.DEBUG)
    name = "ESCO"
    with pytest.raises(SystemExit):
        utils.read_genomes_info("toto.txt", name)
    assert 'Reading given information on your genomes in toto.txt' in caplog.text
    assert ("ERROR: The info file toto.txt that you gave does not exist. "
            "Please provide the right path/name for this file.") in caplog.text
    assert "Ending program" in caplog.text


def test_read_genomes_info_wrongheader(caplog):
    """
    Read lstinfo file and get all genomes information
    When wrong header (not all required columns), exits and appropriate error message
    """
    caplog.set_level(logging.DEBUG)
    name = "ESCO"
    lstinfo_file = os.path.join(DATA_DIR, "test_files", "lstinfo-wrong-header.lst")
    with pytest.raises(SystemExit):
        utils.read_genomes_info(lstinfo_file, name)
    assert ("Reading given information on your genomes in "
            "test/data/annotate/test_files/lstinfo-wrong-header.lst") in caplog.text
    assert ("ERROR: It seems that your info file test/data/annotate/test_files/lstinfo-wrong-header.lst "
            "does not have a header, "
            "or this header does not have, at least, the required columns tab separated: ") in caplog.text
    assert("to_annotate, gsize nb_conts and L90 (in any order).\nEnding program.") in caplog.text


def test_read_genomes_info_noheader(caplog):
    """
    Read lstinfo file and get all genomes information
    When no header, exits and appropriate error message
    """
    caplog.set_level(logging.DEBUG)
    name = "ESCO"
    lstinfo_file = os.path.join(DATA_DIR, "test_files", "lstinfo-no-header.lst")
    with pytest.raises(SystemExit):
        utils.read_genomes_info(lstinfo_file, name)
    assert ("Reading given information on your genomes in "
            "test/data/annotate/test_files/lstinfo-no-header.lst") in caplog.text
    assert ("ERROR: It seems that your info file test/data/annotate/test_files/lstinfo-no-header.lst "
            "does not have a header, "
            "or this header does not have, at least, the required columns tab separated: ") in caplog.text
    assert("to_annotate, gsize nb_conts and L90 (in any order).\nEnding program.") in caplog.text


def test_read_genomes_info_not_int(caplog):
    """
    Read lstinfo file and get all genomes information
    When a nbcont column is not an int for 1 genome, writes error message and ignores the genome
    """
    caplog.set_level(logging.DEBUG)
    name = "ESCO"
    lstinfo_file = os.path.join(DATA_DIR, "test_files", "lstinfo-not-int-nbcont.lst")
    genomes = utils.read_genomes_info(lstinfo_file, name)
    assert ("Reading given information on your genomes in "
            "test/data/annotate/test_files/lstinfo-not-int-nbcont.lst") in caplog.text
    assert 'Found 2 genomes in total' in caplog.text
    assert ("For genome A_H738-and-B2_A3_5, at least one of your columns 'gsize', "
            "'nb_conts' or 'L90' contains a non numeric value. This genome will be ignored") in caplog.text
    exp = {"genome1.fasta":
            ["genome1", "test/data/annotate/genomes/genome1.fasta",
             "test/data/annotate/genomes/genome1.fasta", 800, 6, 5],
           "genome7.fasta":
            ["genome7", "test/data/annotate/genomes/genome7.fasta",
             "test/data/annotate/genomes/genome7.fasta", 79705, 80, 65]}
    assert genomes == exp


def test_read_genomes_info_not_all_filled(caplog):
    """
    Read lstinfo file and get all genomes information
    1 column not filled for at least 1 genome: exits and write appropriate error message
    """
    caplog.set_level(logging.DEBUG)
    name = "ESCO"
    lstinfo_file = os.path.join(DATA_DIR, "test_files", "lstinfo-not-all-filled.lst")
    with pytest.raises(SystemExit):
        utils.read_genomes_info(lstinfo_file, name)
    assert ("Reading given information on your genomes in "
            "test/data/annotate/test_files/lstinfo-not-all-filled.lst") in caplog.text
    assert ("ERROR: Check that all fields of test/data/annotate/test_files/lstinfo-not-all-filled.lst "
            "are filled in each line (can be 'NA')") in caplog.text


def test_read_genomes_info_no_path(caplog):
    """
    Read lstinfo file and get all genomes information
    When a genome in lstinfo does not have its corresponding file, write appropriate error message
    and ignores genome
    """
    caplog.set_level(logging.DEBUG)
    name = "ESCO"
    lstinfo_file = os.path.join(DATA_DIR, "test_files", "lstinfo-miss-1genome.lst")
    genomes = utils.read_genomes_info(lstinfo_file, name)
    assert ("data/annotate/genomes/A_H738-and-B2_A3_5.fna genome file does not exist. "
            "This genome will be ignored") in caplog.text
    assert 'Found 2 genomes in total' in caplog.text
    exp = {"genome1.fasta":
            ["genome1", "test/data/annotate/genomes/genome1.fasta",
             "test/data/annotate/genomes/genome1.fasta", 800, 6, 5],
           "genome7.fasta":
            ["genome7", "test/data/annotate/genomes/genome7.fasta",
             "test/data/annotate/genomes/genome7.fasta", 79705, 80, 65]}
    assert genomes == exp


def test_read_genomes_info_no_genomes(caplog):
    """
    Read lstinfo file and get all genomes information
    When no genome in lstinfo correspond to existing paths, exits and appropriate error message
    """
    caplog.set_level(logging.DEBUG)
    name = "ESCO"
    lstinfo_file = os.path.join(DATA_DIR, "test_files", "lstinfo-no-genome.lst")
    with pytest.raises(SystemExit):
        utils.read_genomes_info(lstinfo_file, name)
    assert ("Reading given information on your genomes in "
            "test/data/annotate/test_files/lstinfo-no-genome.lst") in caplog.text
    assert ("No genome listed in test/data/annotate/test_files/lstinfo-no-genome.lst "
            "was found.") in caplog.text


def test_read_genomes_info_ok(caplog):
    """
    Read lstinfo file and get all genomes information and returns it as expected
    """
    caplog.set_level(logging.DEBUG)
    list_file = os.path.join(DATA_DIR, "test_files", "lstinfo.lst")
    name = "ESCO"
    genomes = utils.read_genomes_info(list_file, name)
    assert ("Reading given information on your genomes in "
            "test/data/annotate/test_files/lstinfo.lst") in caplog.text
    assert 'Found 3 genomes in total' in caplog.text
    exp = {"genome1.fasta":
            ["genome1", "test/data/annotate/genomes/genome1.fasta",
             "test/data/annotate/genomes/genome1.fasta", 800, 6, 6],
           "A_H738-and-B2_A3_5.fna":
            ["A_H738-and-B2_A3_5", "test/data/annotate/genomes/A_H738-and-B2_A3_5.fna",
             "test/data/annotate/genomes/A_H738-and-B2_A3_5.fna", 7000, 78, 5],
           "genome7.fasta":
            ["genome7", "test/data/annotate/genomes/genome7.fasta",
             "test/data/annotate/genomes/genome7.fasta", 79705, 80, 65]}
    assert genomes == exp


def test_read_genomes_info_date_ok(caplog):
    """
    Read lstinfo file and get all genomes information
    """
    caplog.set_level(logging.DEBUG)
    list_file = os.path.join(DATA_DIR, "test_files", "lstinfo.lst")
    name = "ESCO"
    date = "0720"
    genomes = utils.read_genomes_info(list_file, name, date=date)
    assert ("Reading given information on your genomes in "
            "test/data/annotate/test_files/lstinfo.lst") in caplog.text
    assert 'Found 3 genomes in total' in caplog.text
    exp = {"test/data/annotate/genomes/genome1.fasta":
            ["ESCO.0720", "test/data/annotate/genomes/genome1.fasta",
             "test/data/annotate/genomes/genome1.fasta", 800, 6, 6],
           "test/data/annotate/genomes/A_H738-and-B2_A3_5.fna":
            ["ESCO.0720", "test/data/annotate/genomes/A_H738-and-B2_A3_5.fna",
             "test/data/annotate/genomes/A_H738-and-B2_A3_5.fna", 7000, 78, 5],
           "test/data/annotate/genomes/genome7.fasta":
            ["ESCO.0720", "test/data/annotate/genomes/genome7.fasta",
             "test/data/annotate/genomes/genome7.fasta", 79705, 80, 65]}
    assert genomes == exp


def test_check_format():
    """
    test that format is ok or not
    """
    assert utils.check_format("ESC1")
    assert utils.check_format("1234")
    assert not utils.check_format("12")
    assert not utils.check_format("ESC*")


def test_check_resdirlst(caplog):
    """
    Test that when the result directory already contains .lst files in LSTINFO,
    program ends with an error message.
    """
    # Create output directory with a lst file in LSTINFO
    os.makedirs(os.path.join(GENEPATH, "LSTINFO"))
    open(os.path.join(GENEPATH, "LSTINFO", "toto.lst"), "w").close()
    with pytest.raises(SystemExit):
        utils.check_out_dirs(GENEPATH)
    assert ("ERROR: Your output directory already has .lst files in the "
            "LSTINFO folder. Provide another result directory, or remove the "
            "files in this one.") in caplog.text


def test_check_resdirprt(caplog):
    """
    Test that when the result directory already contains .prt files in Proteins,
    program ends with an error message.
    """
    # Create output directory with a lst file in LSTINFO
    os.makedirs(os.path.join(GENEPATH, "Proteins"))
    open(os.path.join(GENEPATH, "Proteins", "toto.prt"), "w").close()
    with pytest.raises(SystemExit):
        utils.check_out_dirs(GENEPATH)
    assert ("ERROR: Your output directory already has .prt files in the "
            "Proteins folder. Provide another result directory, or remove the "
            "files in this one.") in caplog.text


def test_check_resdirgen(caplog):
    """
    Test that when the result directory already contains .gen files in Genes,
    program ends with an error message.
    """
    # Create output directory with a lst file in LSTINFO
    os.makedirs(os.path.join(GENEPATH, "Genes"))
    open(os.path.join(GENEPATH, "Genes", "toto.gen"), "w").close()
    with pytest.raises(SystemExit):
        utils.check_out_dirs(GENEPATH)
    assert ("ERROR: Your output directory already has .gen files in the "
            "Genes folder. Provide another result directory, or remove the "
            "files in this one.") in caplog.text


def test_check_resdirrep(caplog):
    """
    Test that when the result directory already contains .fna files in Replicons,
    program ends with an error message.
    """
    # Create output directory with a lst file in LSTINFO
    os.makedirs(os.path.join(GENEPATH, "Replicons"))
    open(os.path.join(GENEPATH, "Replicons", "toto.fna"), "w").close()
    with pytest.raises(SystemExit):
        utils.check_out_dirs(GENEPATH)
    assert ("ERROR: Your output directory already has .fna files in the "
            "Replicons folder. Provide another result directory, or remove the "
            "files in this one.") in caplog.text


def test_check_resdirgff(caplog):
    """
    Test that when the result directory already contains .fna files in Replicons,
    program ends with an error message.
    """
    # Create output directory with a lst file in LSTINFO
    os.makedirs(os.path.join(GENEPATH, "gff3"))
    open(os.path.join(GENEPATH, "gff3", "toto.gff"), "w").close()
    with pytest.raises(SystemExit):
        utils.check_out_dirs(GENEPATH)
    assert ("ERROR: Your output directory already has .gff files in the "
            "gff3 folder. Provide another result directory, or remove the "
            "files in this one.") in caplog.text


def test_check_resdirotherext():
    """
    Test that when the result directory contains txt files in Replicons dir,
    there is no problem.
    """
    # Create output directory with a lst file in LSTINFO
    os.makedirs(os.path.join(GENEPATH, "Replicons"))
    open(os.path.join(GENEPATH, "Replicons", "toto.txt"), "w").close()
    utils.check_out_dirs(GENEPATH)


def test_check_resdirnodir():
    """
    Test that when the result directory does not already exist, there is no problem.
    """
    utils.check_out_dirs("totoresdir")


def test_run_cmd_error_noquit(caplog):
    """
    Test that when we try to run a command which does not exist, it returns an error message,
    but does not exit the program (eof=False).
    """
    cmd = "toto"
    error = "error trying to run toto"
    assert utils.run_cmd(cmd, error) == 1
    assert "error: command '>toto' is not possible" in caplog.text


def test_run_cmd_error_noquit_logger(caplog):
    """
    Test that when we try to run a command which does not exist, it returns an error message,
    but does not exit the program (eof=False). With a given logger where error is written.
    """
    cmd = "toto"
    logger = logging.getLogger("default")
    error = "error trying to run toto"
    assert utils.run_cmd(cmd, error, logger=logger) == 1
    assert "error: command '>toto' is not possible" in caplog.text


def test_run_cmd_error_quit(caplog):
    """
    Test that when we try to run a command which does not exist, it returns an error message,
    and exits the program (eof=True)
    """
    cmd = "toto"
    error = "error trying to run toto"
    with pytest.raises(SystemExit):
        utils.run_cmd(cmd, error, eof=True)
    assert "command '>toto' is not possible" in caplog.text


def test_run_cmd_retcode_non0(caplog):
    """
    Test that when the command fails, it returns a non-zero int, writes an error message,
    but does not quit the program (eof=False by default).
    """
    cmd = "prokka"
    error = "error trying to run prodigal"
    assert utils.run_cmd(cmd, error).returncode != 0
    assert error in caplog.text


def test_run_cmd_retcode_non0_quit(caplog):
    """
    Test that when the command fails, it returns a non-zero int, writes an error message,
    and exits the program (eof=True).
    """
    cmd = "prokka"
    error = "error trying to run prodigal"
    with pytest.raises(SystemExit):
        utils.run_cmd(cmd, error, eof=True)
    assert error in caplog.text


def test_run_cmd_error_stderrfile(caplog):
    """
    Test that when we try to run a command which does not exist, and direct its output to
    a file instead of stderr, we have the expected error written in the given error file.
    """
    cmd = "toto"
    error = "error trying to run toto"
    outfile = open(os.path.join(GENEPATH, "stderr_run_cmd.txt"), "w")
    assert utils.run_cmd(cmd, error, stderr=outfile) == 1
    assert "error: command '>toto' is not possible" in caplog.text
    outfile.close()


def test_run_cmd_error_stdoutfile(caplog):
    """
    Test that when we try to run a command which does not exist, and direct its output to
    a file instead of stderr, we have the expected error written in the given error file.
    """
    cmd = "toto"
    error = "error trying to run toto"
    outfile = open(os.path.join(GENEPATH, "stdout_run_cmd.txt"), "w")
    assert utils.run_cmd(cmd, error, stdout=outfile) == 1
    assert "error: command '>toto' is not possible" in caplog.text
    outfile.close()


def test_rename_contigs():
    """
    From a given sequence, rename all its contigs with the given gembase name + a number,
    and save the output sequence to the given res_path.
    Check that the output file is as expected.
    """
    logger = logging.getLogger("default")
    gpath = os.path.join(DATA_DIR, "genomes", "H299_H561.fasta")
    gembase_name = "ESCO.0216.00005"
    outfile = os.path.join(GENEPATH, "H299_H561.fasta-short-contig.fna")
    exp_file = os.path.join(DATA_DIR, "exp_files", "res_H299_H561-ESCO00005.fna")
    contigs, sizes = utils.get_genome_contigs_and_rename(gembase_name, gpath, outfile, logger)
    assert contigs == {"H561_S27":"ESCO.0216.00005.0001",
                       "H561_S28":"ESCO.0216.00005.0002",
                       "H561_S29":"ESCO.0216.00005.0003"}
    assert sizes == {"ESCO.0216.00005.0001":3480,
                     "ESCO.0216.00005.0002":7080,
                     "ESCO.0216.00005.0003":2583}
    assert utilities.compare_order_content(outfile, exp_file)


def test_rename_contigs_empty_fasta(caplog):
    """
    From an empty fasta file, ask to rename contigs. Should return empty tuple of dict, and error
    message.
    """
    logger = logging.getLogger("default")
    gpath = os.path.join(GENEPATH, "empty.fasta")
    # Create empty fasta file
    open(gpath, "w").close()
    gembase_name = "ESCO.0216.00005"
    outfile = os.path.join(GENEPATH, "H299_H561.fasta-short-contig.fna")
    exp_file = os.path.join(DATA_DIR, "exp_files", "res_H299_H561-ESCO00005.fna")
    contigs, sizes = utils.get_genome_contigs_and_rename(gembase_name, gpath, outfile, logger)
    assert contigs == {}
    assert sizes == {}
    # Check that outfile is empty
    with open(outfile, "r") as of:
        assert len(of.readlines()) == 0
    # Check error message
    assert ("Your genome test/data/annotate/generated_by_unit-tests/empty.fasta does "
            "not contain any sequence, or is not in fasta format") in caplog.text


def test_rename_contigs_non_fasta(caplog):
    """
    From an file which is not empty, but not in fasta format (no headers starting with '>'),
    check that contigs, sizes and outfile are empty, and error message is ok
    """
    logger = logging.getLogger("default")
    gpath = os.path.join(TEST_DIR, "non-fasta.seq")
    # Create empty fasta file
    open(gpath, "w").close()
    gembase_name = "ESCO.0216.00005"
    outfile = os.path.join(GENEPATH, "H299_H561.fasta-short-contig.fna")
    exp_file = os.path.join(DATA_DIR, "exp_files", "res_H299_H561-ESCO00005.fna")
    contigs, sizes = utils.get_genome_contigs_and_rename(gembase_name, gpath, outfile, logger)
    assert contigs == {}
    assert sizes == {}
    # Check that outfile is empty
    with open(outfile, "r") as of:
        assert len(of.readlines()) == 0
    # Check error message
    assert ("Your genome test/data/annotate/test_files/non-fasta.seq does "
            "not contain any sequence, or is not in fasta format") in caplog.text


def test_rename_contigs_duplicate(caplog):
    """
    From a given sequence, there are 2 contigs named "contig2". Stops and returns false
    """
    logger = logging.getLogger("default")
    gpath = os.path.join(DATA_DIR, "genomes", "genome-duplicated-header.fasta")
    gembase_name = "ESCO.0216.00005"
    outfile = os.path.join(GENEPATH, "genome_dup_error.fna")
    exp_file = os.path.join(DATA_DIR, "exp_files", "res_H299_H561-ESCO00005.fna")
    contigs, sizes = utils.get_genome_contigs_and_rename(gembase_name, gpath, outfile, logger)
    assert not contigs
    assert not sizes
    with open(outfile, "r") as of:
        assert of.readline().startswith(">ESCO.0216.00005.0001")
        of.readline() # skip sequence
        assert of.readline().startswith(">ESCO.0216.00005.0002")
        of.readline() # skip sequence
        assert of.readline().startswith(">ESCO.0216.00005.0003")
    assert ("several contigs have the same name contig2 in test/data/annotate/genomes/"
            "genome-duplicated-header.fasta.") in caplog.text


def test_rename_contigs_duplicate_last(caplog):
    """
    The last contig of the sequence has the same name as a previous contig. Stops and returns false
    """
    logger = logging.getLogger("default")
    gpath = os.path.join(DATA_DIR, "genomes", "genome-duplicated-header-last.fasta")
    gembase_name = "ESCO.0216.00005"
    outfile = os.path.join(GENEPATH, "genome_dup_error.fna")
    exp_file = os.path.join(DATA_DIR, "exp_files", "res_H299_H561-ESCO00005.fna")
    contigs, sizes = utils.get_genome_contigs_and_rename(gembase_name, gpath, outfile, logger)
    assert not contigs
    assert not sizes
    with open(outfile, "r") as of:
        assert of.readline().startswith(">ESCO.0216.00005.0001")
        of.readline() # skip sequence
        assert of.readline().startswith(">ESCO.0216.00005.0002")
        of.readline() # skip sequence
        assert of.readline().startswith(">ESCO.0216.00005.0003")
    assert ("several contigs have the same name contig2 in test/data/annotate/genomes/"
            "genome-duplicated-header-last.fasta.") in caplog.text


def test_cat_nobar():
    """
    Check that when cat is called on a list of several files, the output file
    contains what is expected (concatenation of content of all input files)
    """
    genomes = ["genome1.fasta", "genome_long_header.fst", "genome6.fasta"]
    list_files = [os.path.join(DATA_DIR, "genomes", gen) for gen in genomes]
    outfile = os.path.join(GENEPATH, "test_catfile_nobar.txt")
    utils.cat(list_files, outfile)
    exp_file = os.path.join(DATA_DIR, "exp_files", "res_test_cat-genomes.fst")
    assert utilities.compare_file_content(outfile, exp_file)


def test_cat_bar():
    """
    Check that when cat is called on a list of several files, the output file
    contains what is expected (concatenation of content of all input files)
    """
    genomes = ["genome1.fasta", "genome_long_header.fst", "genome6.fasta"]
    list_files = [os.path.join(DATA_DIR, "genomes", gen) for gen in genomes]
    outfile = os.path.join(GENEPATH, "test_catfile-bar.txt")
    title = "test cat progressbar"
    utils.cat(list_files, outfile, title=title)
    exp_file = os.path.join(DATA_DIR, "exp_files", "res_test_cat-genomes.fst")
    assert utilities.compare_file_content(outfile, exp_file)


def test_detail():
    """
    Check value returned for detail level
    """
    assert utils.detail_lvl() == 15


def test_grep():
    """
    Check that it returns all lines containing the given pattern
    """
    lines = ("1toto.txt otherletters\n"
             "123toto.txtotherletters\n"
             "123toto.txt other letters\n"
             "123toto:txt otherletters\n"
             "123toto.txt otherletters\n")
    filein = os.path.join(GENEPATH, "filein.txt")
    with open(filein, "w") as ff:
        ff.write(lines)
    pattern = "[0-9]+toto[.]txt [a-z]{6}"
    lines_grep = utils.grep(filein, pattern)
    exp = ["1toto.txt otherletters",
           "123toto.txt otherletters"]
    assert exp == lines_grep


def test_grep_count():
    """
    Check that it returns the number of lines containing the given pattern
    """
    lines = ("1toto.txt otherletters\n"
             "123toto.txtotherletters\n"
             "123toto.txt other letters\n"
             "123toto:txt otherletters\n"
             "123toto.txt otherletters\n")
    filein = os.path.join(GENEPATH, "filein.txt")
    with open(filein, "w") as ff:
        ff.write(lines)
    pattern = "[0-9]+toto[.]txt [a-z]{6}"
    lines_grep = utils.grep(filein, pattern, counts=True)
    assert lines_grep == 2


def test_count_lines():
    """
    Test that it returns the number of lines in given file
    """
    lines = ("1toto.txt otherletters\n"
             "123toto.txtotherletters\n"
             "123toto.txt other letters\n"
             "123toto:txt otherletters\n"
             "123toto.txt otherletters\n")
    filein = os.path.join(GENEPATH, "filein.txt")
    with open(filein, "w") as ff:
        ff.write(lines)
    nbline = utils.count(filein)
    assert nbline == 5


def test_count_words():
    """
    Test that it returns the number of words in given file
    """
    lines = ("1toto.txt otherletters\n"
             "123toto.txtotherletters\n"
             "123toto.txt other letters\n"
             "123toto:txt otherletters\n"
             "123toto.txt otherletters\n")
    filein = os.path.join(GENEPATH, "filein.txt")
    with open(filein, "w") as ff:
        ff.write(lines)
    nbword = utils.count(filein, get="words")
    assert nbword == 10


def test_count_error(caplog):
    """
    test that when we want to count something else than 'lines' or 'words', it returns an error
    """
    with pytest.raises(SystemExit):
        utils.count("filein", get="letters")
    assert "Choose what you want to count among ['lines', 'words']" in caplog.text


def test_save_bin():
    """
    Test that python objects are correctly saved into binary file
    """
    obj1 = {1: "toto", "key": 455, "key2": "val"}
    obj2 = [1, 2, 5, "toto", "plop"]
    obj3 = "a string"
    objects = [obj1, obj2, obj3]
    fileout = os.path.join(GENEPATH, "fileout.txt")
    utils.save_bin(objects, fileout)
    import _pickle as pickle
    with open(fileout, "rb") as binf:
        obj = pickle.load(binf)
    assert objects == obj


def test_load_bin():
    """
    Test that python objects are correctly recovered from binary file
    """
    fileout = os.path.join("test", "data", "utils", "objects_saved.bin")
    found = utils.load_bin(fileout)
    obj1 = {1: "toto", "key": 455, "key2": "val"}
    obj2 = [1, 2, 5, "toto", "plop"]
    obj3 = "a string"
    objects = [obj1, obj2, obj3]
    assert found == objects


def test_list_to_str():
    """
    Test that the given list is returned as expected string
    """
    inlist = [1, 2, "toto", {1: 2, "1": 5}, 1e-6]
    tostr = utils.list_to_str(inlist, sep=" ")
    exp1 = "1 2 toto {1: 2, '1': 5} 1e-06\n"
    exp2 = "1 2 toto {'1': 5, 1: 2} 1e-06\n"
    assert tostr == exp1 or tostr == exp2


def test_list_to_str_default():
    """
    Test that the given list is returned as expected string
    """
    inlist = [1, 2, "toto", {1: 2, "1": 5}, 1e-6]
    tostr = utils.list_to_str(inlist)
    exp1 = "1\t2\ttoto\t{1: 2, '1': 5}\t1e-06\n"
    exp2 = "1\t2\ttoto\t{'1': 5, 1: 2}\t1e-06\n"
    assert tostr == exp1 or tostr == exp2


def test_write_list():
    """
    test that given a list, it writes all its elements, 1 by line, in given outfile
    """
    outfile = os.path.join(GENEPATH, "toto.txt")
    inlist = [1, 2, "toto", {1: 2, "1": 5}, 1e-6]
    utils.write_list(inlist, outfile)
    with open(outfile, "r") as of:
        lines = of.readlines()
    assert len(lines) == 5
    assert "1\n" in lines
    assert "2\n" in lines
    assert "toto\n" in lines
    assert "{1: 2, '1': 5}\n" in lines
    assert "1e-06\n" in lines


def test_remove_exits():
    """
    Test that a given file is removed if it exists
    """
    infile = os.path.join(GENEPATH, "toto.txt")
    open(infile, "w").close()
    assert os.path.isfile(infile)
    utils.remove(infile)
    assert not os.path.isfile(infile)


def test_remove_not_exist():
    """
    Test that removing a file which does not exist brings no error
    """
    infile = os.path.join(GENEPATH, "toto.txt")
    assert not os.path.isfile(infile)
    utils.remove(infile)
    assert not os.path.isfile(infile)



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Concatenate all alignment files of all families. Then,
group alignments by genome.

@author: GEM, Institut Pasteur
March 2017
"""

import os
import sys
import logging
import progressbar
import multiprocessing
from PanACoTA import utils

logger = logging.getLogger("align.post")


def post_alignment(fam_nums, all_genomes, prefix, outdir, dname, quiet):
    """
    After the alignment of all proteins by family:

    - concatenate all alignment files
    - group the alignment by genome

    Parameters
    ----------
    fam_nums : []
        list of family numbers
    all_genomes : []
        list of all genomes in dataset
    prefix : str
        path to ``aldir/<name of dataset>`` (used to get extraction, alignment and btr files easily)
    outdir : str
        path to output directory, containing Aldir and Listdir, and that will also contain Treedir
    dname : str
        name of dataset (used to name concat and grouped files, as well as tree folder)
    quiet : bool
        True if nothing must be sent to sdtout/stderr, False otherwise
    """
    all_alns, status = concat_alignments(fam_nums, prefix, quiet)
    treedir = os.path.join(outdir, "Phylo-" + dname)
    os.makedirs(treedir, exist_ok=True)
    outfile = os.path.join(treedir, dname + ".grp.aln")
    res = launch_group_by_genome(all_genomes, all_alns, status, outfile, dname, quiet)
    if not res:
        logger.error("An error occurred. We could not group sequences by genome.")
    return outfile


def concat_alignments(fam_nums, prefix, quiet):
    """
    Concatenate all family alignment files to a unique file

    Parameters
    ----------
    fam_nums : []
        list of family numbers
    prefix : str
        path to ``aldir/<name of dataset>`` (used to get extraction, alignment and btr files easily)
    quiet : bool
        True if nothing must be sent to sdtout/stderr, False otherwise

    Returns
    -------
    tuple
        (output, str) with:

        - output: path to file containing concatenation of all alignments
        - str: "OK" if concatenation file already exists, "Done" if just did concatenation
    """
    output = "{}-complete.cat.aln".format(prefix)
    if os.path.isfile(output):
        logger.info("Alignments already concatenated")
        logger.warning(("Alignments already concatenated in {}. Program will use "
                        "it for next steps. If you want to redo it, remove it before "
                        "running.").format(output))
        return output, "OK"
    logger.info("Concatenating all alignment files")
    list_files = ["{}-mafft-prt2nuc.{}.aln".format(prefix, num_fam) for num_fam in fam_nums]
    # Check that all files exist
    for f in list_files:
        if not os.path.isfile(f):
            logger.error("The alignment file {} does not exist. Please check the families you "
                         "want, and their corresponding alignment files".format(f))
            sys.exit(1)
    if quiet:
        utils.cat(list_files, output)
    else:
        utils.cat(list_files, output, title="Concatenation")
    return output, "Done"


def launch_group_by_genome(all_genomes, all_alns, status, outfile, dname, quiet):
    """
    Function calling group_by_genome in a pool, while giving information to user
    (time elapsed)

    Parameters
    ----------
    all_genomes : []
        list of all genomes in the dataset
    all_alns : str
        path to file containing all alignments concatenated
    status : str
        "OK" if concatenation file already existed before running, "Done" if just did concatenation
    outfile : str
        file containing all families align by genome
    dname : str
        name of dataset
    quiet : bool
        True if nothing must be sent to sdtout/stderr, False otherwise

    Returns
    -------
    bool
        - True if everything went well or was already done
        - False if error occurred in at least one step
    """
    # Status = Done means that we just did the concatenation. So, if grouped by genome
    # file already exists, remove it.
    if status == "Done":
        if os.path.isfile(outfile):
            utils.remove(outfile)
    # Status was not 'Done' (it was 'ok', concat file already existed). And by_genome file
    # also already exists. Warn user
    if os.path.isfile(outfile):
        logger.info("Alignments already grouped by genome")
        logger.warning((f"Alignments already grouped by genome in {outfile}. "
                        "Program will end. "))
        return True
    logger.info("Grouping alignments per genome")
    bar = None
    if not quiet:
        widgets = [progressbar.BouncingBar(marker=progressbar.RotatingMarker(markers="◐◓◑◒")),
                   "  -  ", progressbar.Timer()]
        bar = progressbar.ProgressBar(widgets=widgets, max_value=20, term_width=50)
    pool = multiprocessing.Pool(1)
    args = [all_genomes, all_alns, outfile]
    final = pool.map_async(group_by_genome, [args], chunksize=1)
    pool.close()
    if not quiet:
        while True:
            if final.ready():
                break
            bar.update()
        bar.finish()
    pool.join()
    return False not in final.get()


def group_by_genome(args):
    """
    From the alignment file 'all_alns' containing all proteins, group the alignments of
    proteins by their genome (listed in 'all_genomes'), and save the result
    in 'treedir'

    Parameters
    ----------
    args : tuple
        - all_genomes: list of all genomes
        - all_alns: path to file containing all alignments concatenated
        - outfile: path to file which will contain alignments grouped by genome

    Returns
    -------
    bool
        - True if everything went well
        - False if problem when trying to group by genomes
    """
    all_genomes, all_alns, outfile = args
    sequences = read_alignments(all_alns, all_genomes)
    if not sequences:
        return False
    write_groups(outfile, sequences)
    return True


def read_alignments(all_alns, all_genomes):
    """
    Read alignment file, and assign each sequence to a genome

    Parameters
    ----------
    all_alns : str
        path to file containing all alignments concatenated
    all_genomes : []
        list of all genomes

    Returns
    -------
    dict or None
        - {genome_name: [list of sequences for this genome]}
        - None if problem with a protein for which we don't find the genome
    """
    sequences = {}  # name: [ordered list of sequences]
    genome = None
    seq = ""
    with open(all_alns, 'r') as alnf:
        for line in alnf:
            if line.startswith(">"):
                # If new header, write previous protein name/sequence to 'sequences'
                if genome and seq:
                    sequences[genome].append(seq)
                    seq = ""
                # Get new genome header
                genome = get_genome(line, all_genomes)
                if not genome:
                    return None
                if genome not in sequences:
                    sequences[genome] = []
            else:
                seq += line.strip()
    if genome and seq:
        sequences[genome].append(seq)
    per_genome = [len(seq) for seq in sequences.values()]
    if len(set(per_genome)) != 1:
        logger.error("Problems occurred while grouping alignments by genome: all genomes "
                     "do not have the same number of sequences. Check that each protein "
                     "name contains the name of the genome from which it comes.")
        return None
    logger.details("{} sequences found per genome".format(per_genome[0]))
    return sequences


def write_groups(outfile, sequences):
    """
    Writing alignments per genome to output file.

    Parameters
    ----------
    outfile : str
        path to file that will contain alignments grouped by genome
    sequences : dict
        {genome_name: [list of sequences (DNA, prot...) for this genome]}
    """
    logger.details("Writing alignments per genome")
    with open(outfile, "w") as outf:
        for genome in sorted(sequences, key=utils.sort_genomes_by_name):
            # write header for genome
            outf.write(">" + genome + "\n")
            # Write all sequences
            outf.write("".join(sequences[genome]) + "\n")


def get_genome(header, all_genomes):
    """
    Find to which genome belongs 'header'

    Parameters
    ----------
    header : str
        header read in alignment file
    all_genomes : []
        list of all genomes

    Returns
    -------
    str or None
        name of genome from which the header is
        None if no genome found
    """
    header = header.split(">")[1].split()[0]
    for genome in all_genomes:
        if genome in header:
            return genome
    logger.error((f"Protein {header} does not correspond to any genome name "
                  f"given... {all_genomes}"))
    return None



#!/usr/bin/env python3
# coding: utf-8

"""
Setup script
"""

import PanACoTA
try:
    from setuptools import setup, find_packages
    from setuptools.command.test import test as TestCommand
except ImportError:
    from distutils.core import setup, find_packages
    from distutils.core import Command as TestCommand

class PyTest(TestCommand):
    def initialize_options(self):
        pass

    def finalize_options(self):
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        import sys
        errno = pytest.main(self.test_args)
        sys.exit(errno)


def parse_requirements(requirements):
    """
    Returns the list of requirements found in the requirements file
    """
    with open(requirements, "r") as req_file:
        return [l.strip('\n') for l in req_file if l.strip('\n')
                and not l.startswith('#')]


# packages = ['PanACoTA', 'PanACoTA.prepare_module', 'PanACoTA.annotate_module',
#             'PanACoTA.pangenome_module', 'PanACoTA.corepers_module',
#             'PanACoTA.align_module', 'PanACoTA.tree_module', 'PanACoTA.subcommands']
requires = parse_requirements("requirements.txt")
scripts = ['PanACoTA/bin/run_panacota.py']

classifiers = [
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: GNU Affero General Public License v3",
    "Programming Language :: Python :: 3",
    "Operating System :: OS Independent",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
]

with open('README.md') as f:
    long_description = f.read()

setup(
    name='PanACoTA',
    packages=find_packages(),
    version=PanACoTA.__version__,
    description="Large scale comparative genomics tools: annotate genomes, do pangenome, "
                "core/persistent genome, align core/persistent families, infer phylogenetic tree.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Amandine Perrin',
    author_email='amandine.perrin@pasteur.fr',
    license='AGPL v3',
    platforms='OS Independent',
    package_data={'': ['LICENSE']},
    url='https://github.com/gem-pasteur/PanACoTA',
    scripts=scripts,
    entry_points={
          'console_scripts': ['PanACoTA=PanACoTA.bin.run_panacota:main']
      },
    python_requires='>=3.4',
    include_package_data=True,
    install_requires=requires,
    extras_require={'dev': parse_requirements("requirements-dev.txt")},
    tests_require=['pytest'],
    cmdclass={'test': PyTest},
    classifiers=classifiers
)



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
'all' is a module of PanACoTA, allowing to run the whole pipeline at once.


@author gem
October 2020
"""

import os
import sys
from termcolor import colored
import sys

from PanACoTA import utils
from PanACoTA import utils_argparse
from PanACoTA.subcommands import prepare
from PanACoTA.subcommands import annotate
from PanACoTA.subcommands import pangenome
from PanACoTA.subcommands import corepers
from PanACoTA.subcommands import align
from PanACoTA.subcommands import tree
from PanACoTA import __version__ as version


def main_from_parse(args):
    """
    Call main function from the arguments given by parser

        verbosity:

    - defaut 0 : stdout contains INFO, stderr contains ERROR.
    - 1: stdout contains INFO, stderr contains WARNING and ERROR
    - 2: stdout contains (DEBUG), DETAIL and INFO, stderr contains WARNING and ERROR
    - >=15: Add DEBUG in stdout

    Parameters
    ----------
    args_all : tuple
        arguments common to all modules: output directory (str),
        threads (int), verbose (int), quiet (bool)
    args_prepare : tuple
        arguments for prepare module (see subcommands.prepare.py): NCBI_species_taxid (int),
        NCBI_species (str), levels (str), tmp_dir (str), norefseq (bool), db_dir (str),
        only_mash (bool), info_file (str), l90 (int), nbcont (int), cutn (int),
        min_dist (float), max_dist (float)
    args_annot : tuple
        arguments for annotate module (see subcommands/annotate.py): name (str), qc_only (bool),
        date (str), prodigal_only (bool), small (bool)
    args_pan : tuple
        arguments for pangenome module (see subcommands/pangenome.py): min_id (float),
        clust_mode (int), spe_dir (str), outfile (str)
    args_corepers : tuple
        arguments for corepers module (see subcommands.corepers.py): tol (float), mixed (bool),
        multi (bool), floor (bool)
    args_tree : tuple
        arguments for tree module (see subcommands.tree.py): soft (str), model (str), boot (bool),
        write_boot (bool), memory (str), fast (bool)
    """
    cmd = "PanACoTA " + ' '.join(args.argv)
    args_all = (args.outdir, args.threads, args.verbose, args.quiet)
    args_prepare = (args.ncbi_species_taxid, args.ncbi_species_name, args.ncbi_taxid, args.levels,
                    args.ncbi_section, args.tmp_dir, args.norefseq, args.db_dir, args.only_mash, 
                    args.info_file, args.l90, args.nbcont, args.cutn, args.min_dist, args.max_dist)
    args_annot = (args.name, args.qc_only, args.date, args.prodigal_only, args.small)
    args_pan = (args.min_id, args.clust_mode, args.spedir, args.outfile)
    args_cp = (args.tol, args.mixed, args.multi, args.floor)
    args_tree = (args.soft, args.model, args.boot, args.write_boot, args.memory, args.fast)
    main(cmd, args_all, args_prepare, args_annot, args_pan, args_cp, args_tree)


def main(cmd, args_all, args_prepare, args_annot, args_pan, args_corepers, args_tree):
    """
    Call all modules, one by one, using output of one as input for the next one

    Parameters
    ----------
    cmd : str
        command line used to launch the program
    args_all : tuple
        arguments common to all modules: output directory (str),
        threads (int), verbose (int), quiet (bool)
    args_prepare : tuple
        arguments for prepare module (see subcommands.prepare.py): NCBI_species_taxid (int),
        NCBI_species_name (str), NCBI_taxid (int), levels (str), NCBI_section (str),
        tmp_dir (str), norefseq (bool), db_dir (str),
        only_mash (bool), info_file (str), l90 (int), nbcont (int), cutn (int),
        min_dist (float), max_dist (float)
    args_annot : tuple
        arguments for annotate module (see subcommands/annotate.py): name (str), qc_only (bool),
        date (str), prodigal_only (bool), small (bool)
    args_pan : tuple
        arguments for pangenome module (see subcommands/pangenome.py): min_id (float),
        clust_mode (int), spe_dir (str), outfile (str)
    args_corepers : tuple
        arguments for corepers module (see subcommands.corepers.py): tol (float), mixed (bool),
        multi (bool), floor (bool)
    args_tree : tuple
        arguments for tree module (see subcommands.tree.py): soft (str), model (str), boot (bool),
        write_boot (bool), memory (str), fast (bool)
    """
    outdir, threads, verbose, quiet = args_all
    os.makedirs(outdir, exist_ok=True)
    # Initialize logger
    import logging
    # set level of logger: level is the minimum level that will be considered.
    if verbose <= 1:
        level = logging.INFO
    # for verbose = 2, ignore only debug
    if verbose >= 2 and verbose < 15:
        level = utils.detail_lvl() # int corresponding to detail level
    # for verbose >= 15, write everything
    if verbose >= 15:
        level = logging.DEBUG
    logfile_base = os.path.join(outdir, "PanACoTA-all_modules")
    logfile_base = utils.init_logger(logfile_base, level, name='all_modules',
                                     verbose=verbose, quiet=quiet)
    logger = logging.getLogger('all_modules')
    logger.info(f'PanACoTA version {version}')
    logger.info("Command used\n \t > " + cmd)

    # Run prepare module
    outdir_prepare = os.path.join(outdir, "1-prepare_module")
    (NCBI_species_taxid, NCBI_species_name, NCBI_taxid, levels, NCBI_section,
     tmp_dir, norefseq, db_dir, only_mash, info_file,
     l90, nbcont, cutn, min_dist, max_dist) = args_prepare
    logger.info("prepare step")
    info_file = prepare.main("PanACoTA prepare", NCBI_species_name, NCBI_species_taxid,
                             NCBI_taxid, levels, NCBI_section,
                             outdir_prepare, tmp_dir, threads, norefseq, db_dir, only_mash,
                             info_file, l90, nbcont, cutn, min_dist, max_dist, verbose, quiet)

    # Run annotate module
    list_file = ""
    db_path = ""
    tmp_dir = ""
    force = False
    outdir_annotate = os.path.join(outdir, "2-annotate_module")
    (name, qc_only, date, prodigal_only, small) = args_annot
    res_annot_dir = None

    logger.info("annotate step")
    lstinfo, nbgenomes = annotate.main("PanACoTA annotate", list_file, db_path, outdir_annotate,
                                       name, date, l90, nbcont, cutn, threads, force, qc_only,
                                       info_file, tmp_dir, res_annot_dir, verbose, quiet,
                                       prodigal_only=prodigal_only, small=small)
    if qc_only:
        return "QC_only done"

    # Pangenome step
    name_pan = f"{name}_{nbgenomes}"
    outdir_pan = os.path.join(outdir, "3-pangenome_module")
    dbpath = os.path.join(outdir_annotate, "Proteins")
    (min_id, clust_mode, spe_dir, outfile) = args_pan
    logger.info("pangenome step")
    panfile = pangenome.main("PanACoTA pangenome", lstinfo, name_pan, dbpath, min_id, outdir_pan,
                             clust_mode, spe_dir, threads, outfile, verbose=verbose,
                             quiet=quiet)

    # Coregenome step
    outdir_corpers = os.path.join(outdir, "4-corepers_module")
    logger.info("corepers step")
    (tol, mixed, multi, floor) = args_corepers
    corepers_file = corepers.main("PanACoTA corepers", panfile, tol, multi, mixed, outdir_corpers,
                                  floor, verbose, quiet)
    # Align step
    outdir_align = os.path.join(outdir, "5-align_module")
    force = False
    logger.info("align step")
    align_file = align.main("PanACoTA align", corepers_file, lstinfo, name_pan, outdir_annotate,
                            outdir_align, threads, force, verbose=verbose, quiet=quiet)


    # Tree step
    (soft, model, boot, write_boot, memory, fast) = args_tree
    outdir_tree = os.path.join(outdir, "6-tree_module")
    logger.info("tree step")
    tree.main("PanACoTA tree", align_file, outdir_tree, soft, model, threads, boot,
              write_boot, memory, fast, verbose=verbose, quiet=quiet)
    logger.info("All modules of PanACOTA are finished.")
    return 0


def build_parser(parser):
    """
    Method to create a parser for command-line options

    Parameters
    ----------
    parser : argparse.ArgumentParser
        The parser to configure

    """
    import argparse
    from PanACoTA import utils_argparse

    # Create command-line parser for all options and arguments to give
    general = parser.add_argument_group("General arguments")
    general.add_argument("-c", dest="configfile",
                          help=("Path to your configuration file, defining values of parameters.")
                          )
    general.add_argument("-o", dest="outdir", required=True,
                          help=("Path to your output folder, where all results "
                                "from all 6 modules will be saved.")
                          )
    general.add_argument("--threads", dest="threads", type=utils_argparse.thread_num,
                          help="Specify how many threads can be used (default=1)")
    # prepare arguments
    pprepare = parser.add_argument_group("'prepare' module arguments")
    pprepare.add_argument("-t", dest="ncbi_species_taxid",
                          help=("Species taxid to download, corresponding to the "
                                "'species taxid' provided by the NCBI. A comma-separated "
                                "list of taxid can also be provided.")
                         )
    pprepare.add_argument("-s", dest="ncbi_species",
                          help=("Species to download, corresponding to the "
                                "'organism name' provided by the NCBI. Give name between "
                                "quotes (for example \"escherichia coli\")")
                        )
    pprepare.add_argument("-l", "--assembly_level", dest="levels",
                          help=("Assembly levels of genomes to download (default: all). "
                                "Possible levels are: 'all', 'complete', 'chromosome', "
                                "'scaffold', 'contig'."
                                "You can also provide a comma-separated list of assembly "
                                "levels. For ex: 'complete,chromosome'")
                          )
    pprepare_annote = parser.add_argument_group("Common arguments to 'prepare' "
                                               "and 'annotate' modules")
    pprepare_annote.add_argument("--cutn", dest="cutn", type=utils_argparse.positive_int,

                          help=("By default, each genome will be cut into new contigs when "
                                "at least 5 'N' in a row are found in its sequence. "
                                "If you don't want to "
                                "cut genomes into new contigs when there are rows of 'N', "
                                "put 0 to this option. If you want to cut from a different number "
                                "of 'N' in a row, put this value to this option.")
                          )
    pprepare_annote.add_argument("--l90", dest="l90", type=int,
                                help=("Maximum value of L90 allowed to keep a genome. "
                                      "Default is 100.")
                                )
    pprepare_annote.add_argument("--nbcont", dest="nbcont", type=utils_argparse.cont_num,
                                help=("Maximum number of contigs allowed to "
                                                   "keep a genome. Default is 999."))

    pannote = parser.add_argument_group("'annotate' module arguments")
    pannote.add_argument("--prodigal", dest="prodigal_only", action="store_true",
                        help="Add this option if you only want syntactical annotation, given "
                             "by prodigal, and not functional annotation requiring prokka and "
                             "is slower.")
    pannote.add_argument("-n", dest="name", required=True, type=utils_argparse.gen_name,
                        help=("Choose a name for your annotated genomes. This name should "
                              "contain 4 alphanumeric characters. Generally, they correspond "
                              "to the 2 first letters of genus, and 2 first letters of "
                              "species, e.g. ESCO for Escherichia Coli."))

    ppangenome = parser.add_argument_group("'pangenome' module arguments")
    ppangenome.add_argument("-i", dest="min_id", type=utils_argparse.perc_id,
                           help=("Minimum sequence identity to be considered in the same "
                                 "cluster (float between 0 and 1). Default is 0.8."))

    pcorepers = parser.add_argument_group("'corepers' module arguments")
    pcorepers.add_argument("--tol", dest="tol", type=utils_argparse.percentage,
                          help=("min %% of genomes having at least 1 member in a family to "
                                "consider the family as persistent (between 0 and 1, "
                                "default is 1 = 100%% of genomes = Core genome)."
                                "By default, the minimum number of genomes will be "
                                "ceil('tol'*N) (N being the total number of genomes). If "
                                "you want to use floor('tol'*N) instead, add the '-F' option."))
    pcorepers.add_argument("-Mu", dest="multi", action='store_true',
                          help=("Add this option if you allow several members in any genome "
                                "of a family. By default, only 1 (or 0 if tol<1) member "
                                "per genome are allowed in all genomes. If you want to allow "
                                "exactly 1 member in 'tol'%% of the genomes, and 0, 1 "
                                "or several members in the '1-tol'%% left, use the option -X "
                                "instead of this one: -M and -X options are not compatible."))
    pcorepers.add_argument("-X", dest="mixed", action='store_true',
                          help="Add this option if you want to allow families having several "
                               "members only in '1-tol'%% of the genomes. In the other genomes, "
                               "only 1 member exactly is allowed. This option is not compatible "
                               "with -M (which is allowing multigenic families: having several "
                               "members in any number of genomes).")

    ptree = parser.add_argument_group("'tree' module arguments")
    softs = ["fasttree", "fastme", "quicktree", "iqtree", "iqtree2"]
    ptree.add_argument("--soft", dest="soft", choices=softs,
                      help=("Choose with which software you want to infer the "
                            "phylogenetic tree. Default is IQtree."))

    helper = parser.add_argument_group('Others')
    helper.add_argument("-v", "--verbose", dest="verbose", action="count",
                        help="Increase verbosity in stdout/stderr.")
    helper.add_argument("-q", "--quiet", dest="quiet", action="store_true",
                        help=("Do not display anything to stdout/stderr. log files will "
                              "still be created."))
    helper.add_argument("-h", "--help", dest="help", action="help",
                        help="show this help message and exit")


def parse(parser, argu):
    """
    parse arguments given to parser

    Parameters
    ----------
    parser : argparse.ArgumentParser
        the parser used
    argu : [str]
        command-line given by user, to parse using parser

    Returns
    -------
    argparse.Namespace
        Parsed arguments
    """
    import argparse
    args = parser.parse_args(argu)
    return check_args(parser, args)


def check_args(parser, argv):
    """
    Check arguments given by user

    Parameters
    ----------
    parser : argparse.ArgumentParser
        the parser used
    argv : argparse.Namespace
        parsed command-line given by user

    Returns
    -------
    argv : argparse.Namespace
        parsed command-line 
    """
    from argparse import Namespace

    # Get arguments given in command line
    # If an argument was not in the user command-line, it still exists in argv,
    # but is set to 'None' -> ignore it
    # If a bool argument (quiet, mixed, multi, prodigal_only) is not in the user command-line,
    # it still exists in argv but is set to False -> ignore it to replace by info from
    # config file if any, or default otherwise
    dict_argv = {key:val for key,val in vars(argv).items() if val is not None and val != False}
    final_dict = {}

    # PREPARE STEP
    prep_dict = get_prepare(dict_argv)
    a = Namespace(**prep_dict)
    prepare.check_args(parser, a)
    final_dict.update(prep_dict)

    # ANNOTATE STEP
    annot_dict = get_annotate(dict_argv)
    a = Namespace(**annot_dict)
    annotate.check_args(parser, a)
    final_dict.update(annot_dict)

    # PANGENOME STEP
    pan_dict = get_pangenome(dict_argv)
    final_dict.update(pan_dict)
    # Add default arguments if not found in commandline nor config file

    # COREPERS STEP
    cp_dict = get_corepers(dict_argv)
    a = Namespace(**cp_dict)
    corepers.check_args(parser, a)
    final_dict.update(cp_dict)

    # TREE STEP
    tree_dict = get_tree(dict_argv)
    # If we chose soft==quicktree, check_args will return error if threads > 1. So,
    # we put threads = 1 during check_args, and put back the value after
    th_save = tree_dict["threads"]
    a = Namespace(**tree_dict)
    checked_a = tree.check_args(parser, a)
    # Variable can be changed by check_args (like put default model if not given)
    tree_dict.update(vars(checked_a))
    tree_dict["threads"] = th_save
    final_dict.update(tree_dict)

    # Combine files
    # Put new args values in argv
    for key, val in final_dict.items():
        vars(argv)[key] = val
    return argv


def get_prepare(dict_argv):
    """
    Check that arguments given for prepare step are compatible
    """
    # Get arguments from config file and add them (overwrite if needed)
    if not "configfile" in dict_argv:
        conf_conffile = utils_argparse.Conf_all_parser("",["prepare"])
    else:
        conf_conffile = utils_argparse.Conf_all_parser(dict_argv['configfile'],
                                                       readsec=["prepare"])
    # Add arguments from commandline
    conf_conffile.update(dict_argv, "prepare")
    # Add default arguments if not found in comd line nor config file
    defaults = {"verbose": 0, "threads": 1, "cutn": 5, "l90": 100, "nbcont":999,
                "levels": "all", "quiet": False, "ncbi_species_name": "",
                "ncbi_species_taxid": "", "ncbi_taxid": "", "tmp_dir": "", "db_dir": "",
                "info_file": "", "min_dist": 1e-4, "max_dist": 0.06,
                "norefseq": False, "only_mash": False, "ncbi_section": "refseq"}
    conf_conffile.add_default(defaults, "prepare")
    # Change to expected types (boolean, int, float)
    conf_conffile.set_boolean("prepare", "quiet")
    conf_conffile.set_boolean("prepare", "only_mash")
    conf_conffile.set_boolean("prepare", "norefseq")
    conf_conffile.set_int("prepare", "threads")
    conf_conffile.set_int("prepare", "verbose")
    conf_conffile.set_int("prepare", "cutn")
    conf_conffile.set_int("prepare", "l90")
    conf_conffile.set_int("prepare", "nbcont")
    conf_conffile.set_float("prepare", "min_dist")
    conf_conffile.set_float("prepare", "max_dist")
    prep_dict = conf_conffile.get_section_dict("prepare")
    return prep_dict


def get_annotate(dict_argv):
    """
    Check that arguments given for annotate step are compatible
    """
    # Get arguments from config file and add them (overwrite if needed)
    if not "configfile" in dict_argv:
        conf_conffile = utils_argparse.Conf_all_parser("",["annotate"])
    else:
        conf_conffile = utils_argparse.Conf_all_parser(dict_argv['configfile'],
                                                       readsec=["annotate"])
    # Add arguments from commandline
    conf_conffile.update(dict_argv, "annotate")
    if "date" not in dict(conf_conffile["annotate"]):
        import time
        date = time.strftime("%m%y")
        conf_conffile.update({"date": date}, "annotate")
    # Add default arguments if not found in commandline nor config file
    defaults = {"verbose": 0, "threads": 1, "cutn": 5, "l90": 100, "nbcont":999,
                "quiet": False, "prodigal_only": False, "small": False, "qc_only": False,
                "list_file": "", "db_path": "", "from_info": True}
    conf_conffile.add_default(defaults, "annotate")
    conf_conffile.set_boolean("annotate", "quiet")
    conf_conffile.set_boolean("annotate", "prodigal_only")
    conf_conffile.set_boolean("annotate", "small")
    conf_conffile.set_boolean("annotate", "qc_only")
    conf_conffile.set_int("annotate", "verbose")
    conf_conffile.set_int("annotate", "threads")
    conf_conffile.set_int("annotate", "cutn")
    conf_conffile.set_int("annotate", "nbcont")
    conf_conffile.set_int("annotate", "l90")
    annot_dict = conf_conffile.get_section_dict("annotate")
    return annot_dict


def get_pangenome(dict_argv):
    """
    Check that arguments given for pangenome step are compatible
    """
    # Get arguments from config file and add them (overwrite if needed)
    if not "configfile" in dict_argv:
        conf_conffile = utils_argparse.Conf_all_parser("",["pangenome"])
    else:
        conf_conffile = utils_argparse.Conf_all_parser(dict_argv['configfile'],
                                                       readsec=["pangenome"])
    # Add arguments from commandline
    conf_conffile.update(dict_argv, "pangenome")
    # Add default arguments if not found in commandline nor config file
    defaults = {"verbose": 0, "threads": 1, "min_id": 0.8, "quiet": False, "clust_mode": 1,
                "outfile": "", "spedir": ""}
    conf_conffile.add_default(defaults, "pangenome")
    conf_conffile.set_boolean("pangenome", "quiet")
    conf_conffile.set_int("pangenome", "verbose")
    conf_conffile.set_int("pangenome", "threads")
    conf_conffile.set_float("pangenome", "min_id")
    pan_dict = conf_conffile.get_section_dict("pangenome")
    return pan_dict


def get_corepers(dict_argv):
    """
    Check that arguments given for corepers step are compatible
    """
    # Get arguments from config file and add them (overwrite if needed)
    if not "configfile" in dict_argv:
        conf_conffile = utils_argparse.Conf_all_parser("",["corepers"])
    else:
        conf_conffile = utils_argparse.Conf_all_parser(dict_argv['configfile'],
                                                       readsec=["corepers"])
    # Add arguments from commandline
    conf_conffile.update(dict_argv, "corepers")
    # Add default arguments if not found in commandline nor config file
    defaults = {"verbose": 0, "quiet": False, "tol": 1, "mixed": False, "multi": False,
                "floor": False, "threads": 1}
    conf_conffile.add_default(defaults, "corepers")
    conf_conffile.set_boolean("corepers", "quiet")
    conf_conffile.set_boolean("corepers", "floor")
    conf_conffile.set_boolean("corepers", "mixed")
    conf_conffile.set_boolean("corepers", "multi")
    conf_conffile.set_int("corepers", "verbose")
    conf_conffile.set_int("corepers", "tol")
    conf_conffile.set_int("corepers", "threads")
    corepers_dict = conf_conffile.get_section_dict("corepers")
    return corepers_dict


def get_tree(dict_argv):
    """
    Check that arguments given for tree step are compatible
    """
    # Get arguments from config file and add them (overwrite if needed)
    if not "configfile" in dict_argv:
        conf_conffile = utils_argparse.Conf_all_parser("",["tree"])
    else:
        conf_conffile = utils_argparse.Conf_all_parser(dict_argv['configfile'], readsec=["tree"])
    # Add arguments from commandline
    conf_conffile.update(dict_argv, "tree")
    # Add default arguments if not found in commandline nor config file
    defaults = {"verbose": 0, "quiet": False, "threads": 1,
                "soft": "iqtree", "model": None, "boot": False, "write_boot": False,
                "memory": None, "fast": False}
    conf_conffile.add_default(defaults, "tree")
    conf_conffile.set_boolean("tree", "quiet")
    conf_conffile.set_boolean("tree", "fast")
    conf_conffile.set_boolean("tree", "boot")
    conf_conffile.set_boolean("tree", "write_boot")
    conf_conffile.set_int("tree", "verbose")
    conf_conffile.set_int("tree", "threads")
    tree_dict = conf_conffile.get_section_dict("tree")
    return tree_dict


if __name__ == '__main__':
    import argparse
    from textwrap import dedent
    header = '''
     ___                 _____  ___         _____  _____
    (  _`\              (  _  )(  _`\      (_   _)(  _  )
    | |_) )  _ _   ___  | (_) || ( (_)   _   | |  | (_) |
    | ,__/'/'_` )/' _ `\|  _  || |  _  /'_`\ | |  |  _  |
    | |   ( (_| || ( ) || | | || (_( )( (_) )| |  | | | |
    (_)   `\__,_)(_) (_)(_) (_)(____/'`\___/'(_)  (_) (_)


       Large scale comparative genomics tools

     -------------------------------------------
     '''
    my_parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                        description=dedent(header), add_help=False)
    build_parser(my_parser)
    OPTIONS = parse(my_parser, sys.argv[1:])
    main_from_parse(OPTIONS)



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Function to infer a phylogenetic tree from a persistent genome, using IQtree

@author gem
July 2020
"""

import os
import logging

from PanACoTA import utils

logger = logging.getLogger("tree.iqtree")

def run_tree(alignfile, boot, outdir, quiet, threads, **kwargs):
    """
    Run IQtree for the given alignment file and options

    Parameters
    ----------
    alignfile: str
        path to file containing all persistent families aligned, and grouped by genome
    boot: int or None
        number of bootstraps to calculate, None if no bootstrap asked
    outdir: str or None
        Path to the tree file that must be created
    quiet: bool
        True if nothing must be printed to stderr/stdout, False otherwise
    threads: int
        Maximum number of threads to use
    kwargs["model"]: str
        DNA substitution model chosen by user
    kwards["wb"]: bool
    	True if all bootstrap pseudo-trees must be saved into a file, False otherwise
    kwargs["mem"]: str
    	Maximal RAM usage in GB | MB | % - Only for iqtree
    kwargs["s"]: str
    	soft to use (iqtree or iqtree2)
    """
    # Get optional arguments
    model = kwargs["model"]
    write_boot = kwargs["wb"]
    memory = kwargs["mem"]
    soft = kwargs["s"]
    fast = kwargs["f"]
    if not fast:
        fast = ""
    else:
        fast = "-fast"

    logger.info("Running IQtree...")

    # Init non mandatory arguments
    bootinfo = ""
    wb_info = ""
    mem_info = ""
    threadinfo = ""

    # Get info on all options (syntax changes according to IQtree version 1.x or 2.x)
    if boot:
        if soft=="iqtree":
            bootinfo = f"-bb {boot}"
        else:
            bootinfo = f"-B {boot}"
    if write_boot:
        if soft == "iqtree":
            wb_info = "-wbt"
        else:
    	    wb_info = "--boot-trees"
    if memory:
        if soft=="iqtree":
    	    mem_info = f"-mem {memory}"
        else:
            mem_info = f"--mem {memory}"
    # IQtree is always run quietly, but syntax depends on version:
    if soft=="iqtree":
        qu = "-quiet"
    else:
        qu = "--quiet"
    # Get threads information
    if threads:
    	if soft == "iqtree":
    		threadinfo = f"-nt {threads}"
    	else:
    		threadinfo = f"-T {threads}"

	# get cmd for seqtype
    if soft == "iqtree":
        seqtype = "-st DNA"
    else:
        seqtype = "--seqtype DNA"

    # Define treefile name if not given.
    align_name = os.path.basename(alignfile)
    logfile = os.path.join(outdir, align_name + ".iqtree.log")
    treefile = os.path.join(outdir, align_name + ".iqtree_tree")
    # get prefix cmd:
    if soft == "iqtree":
        prefix = f"-pre {treefile}"
    else:
        prefix = f"--prefix {treefile}"
    cmd = (f"{soft} -s {alignfile} {threadinfo} -m {model} {mem_info} {bootinfo} {wb_info} "
    	   f"{seqtype} {prefix} {qu} {fast}")
    logger.details("IQtree command: " + cmd)
    if quiet:
        fnull = open(os.devnull, 'w')
    else:
        fnull = None
    error = (f"Problem while running IQtree. See log file ({logfile}) for "
             "more information.")
    utils.run_cmd(cmd, error, eof=True, logger=logger, stderr=fnull)



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
align is a subcommand of PanACoTA


@author gem
June 2017
"""

import os
import sys


def main_from_parse(args):
    """
    Call main function from the arguments given by parser

    Parameters
    ----------
    args : argparse.Namespace
        result of argparse parsing of all arguments in command line
    """
    cmd = "PanACoTA " + ' '.join(args.argv)
    main(cmd, args.corepers, args.list_genomes, args.dataset_name, args.dbpath,
         args.outdir, args.threads, args.force, args.verbose, args.quiet)


def main(cmd, corepers, list_genomes, dname, dbpath, outdir, threads, force, verbose=0,
         quiet=False):
    """
    Align given core genome families

    Parameters
    ----------
    corepers : str
        File containing persistent genome families
    list_genomes : str
        File containing the list of all genomes in the dataset. Only first column is
        considered.
    dname : str
        Dataset name, used to name output files
    dbpath : str
        path to the directory containing 'Proteins' and 'Genes' folders
    outdir : str
        path to the directory where output files must be saved
    threads : int
        Max number of threads to use
    force : bool
        Remove existing output files and rerun everything if True.
    verbose : int
        verbosity:
        - defaut 0 : stdout contains INFO, stderr contains ERROR.
        - 1: stdout contains INFO, stderr contains WARNING and ERROR
        - 2: stdout contains (DEBUG), DETAIL and INFO, stderr contains WARNING and ERROR
        - >=15: Add DEBUG in stdout

    quiet : bool
        True if nothing must be sent to stdout/stderr, False otherwise
    """
    # import needed packages
    import logging
    import shutil
    from PanACoTA import utils
    from PanACoTA.align_module import pan_to_pergenome as p2g
    from PanACoTA.align_module import get_seqs as gseqs
    from PanACoTA.align_module import alignment as ali
    from PanACoTA.align_module import post_align as post
    from PanACoTA import __version__ as version

    # test if prokka is installed and in the path
    if not utils.check_installed("mafft"):  # pragma: no cover
        print("mafft is not installed. 'PanACoTA align' cannot run.")
        sys.exit(1)

    if force and os.path.isdir(outdir):
        shutil.rmtree(outdir)
    os.makedirs(outdir, exist_ok=True)
    # set level of logger (here debug to show everything during development)
    # level is the minimum level that will be considered.
    # for verbose = 0 or 1, ignore details and debug, start from info
    if verbose <= 1:
        level = logging.INFO
    # for verbose = 2, ignore only debug
    if verbose >= 2 and verbose < 15:
        level = 15 # int corresponding to detail level
    # for verbose >= 15, write everything
    if verbose >= 15:
        level = logging.DEBUG
    # name logfile, add timestamp if already existing
    logfile_base = os.path.join(outdir, "PanACoTA-align_" + dname)
    utils.init_logger(logfile_base, level, 'align', log_details=True, verbose=verbose, quiet=quiet)
    logger = logging.getLogger("align")
    logger.info(f'PanACoTA version {version}')
    logger.info("Command used\n \t > " + cmd)

    all_genomes, aldir, listdir, fam_nums = p2g.get_per_genome(corepers, list_genomes,
                                                               dname, outdir)
    # generate required files
    gseqs.get_all_seqs(all_genomes, dname, dbpath, listdir, aldir, fam_nums, quiet)
    prefix = os.path.join(aldir, dname)

    # Align all families
    status = ali.align_all_families(prefix, fam_nums, len(all_genomes), dname, quiet, threads)
    if not status:
        logger.error(("At least one alignment did not run well. See detailed log file for "
                      "more information. Program will stop here, alignments won't be "
                      "grouped by genome."))
        sys.exit(1)

    # post-process alignment files
    align_file = post.post_alignment(fam_nums, all_genomes, prefix, outdir, dname, quiet)
    logger.info("END")
    return align_file


def build_parser(parser):
    """
    Method to create a parser for command-line options

    Parameters
    ----------
    parser : argparse.ArgumentParser
        parser to configure in order to extract command-line arguments
    """
    import argparse
    import multiprocessing
    from PanACoTA import utils_argparse

    # Create command-line parser for all options and arguments to give
    required = parser.add_argument_group('Required arguments')
    required.add_argument("-c", dest="corepers", required=True,
                          help="Core or persistent genome whose families must be aligned.")
    required.add_argument("-l", dest="list_genomes", required=True,
                          help=("File containing the list of all the genomes you want "
                                "to align from their core/persistent families. "
                                "1 genome per line: it can be the "
                                "LSTINFO-<list_file>.lst file of 'PanACoTA annotate' module. "
                                "Here, only the first column (genome name without extension) "
                                "will be used. The final alignment file will contain "
                                "1 alignment per genome in this file."))
    required.add_argument("-n", dest="dataset_name", required=True,
                          help=("Name of the dataset which will be aligned (for example, "
                                "SAEN1234 for 1234 Salmonella enterica genomes). This name will "
                                "be used to name the alignment file."))
    required.add_argument("-d", dest="dbpath", required=True,
                          help=("Path to the folder containing the directories 'Proteins' "
                                "and 'Genes', created by 'PanACoTA annotate'."))
    required.add_argument("-o", dest="outdir", required=True,
                          help="Output directory, where all results must be saved ")

    optional = parser.add_argument_group('Optional arguments')
    optional.add_argument("--threads", dest="threads", default=1, type=utils_argparse.thread_num,
                          help=("add this option if you want to parallelize on several threads. "
                                "Indicate on how many threads you want to parallelize. "
                                "By default, it uses 1 thread. Put 0 if you want to use "
                                "all threads of your computer."))
    optional.add_argument("-F", "--force", dest="force", action="store_true",
                          help=("Force run: Add this option if you want to redo all alignments "
                                "for all families, even if their result file already exists. "
                                "Without this option, if an alignment file already exists, "
                                "it will be used for the next step. If you want to redo only "
                                "a given alignment, just delete its file, without using "
                                "this option."))
    helper = parser.add_argument_group('Others')
    helper.add_argument("-v", "--verbose", dest="verbose", action="count", default=0,
                        help="Increase verbosity in stdout/stderr.")
    helper.add_argument("-q", "--quiet", dest="quiet", action="store_true", default=False,
                        help=("Do not display anything to stdout/stderr. log files will "
                              "still be created."))
    helper.add_argument("-h", "--help", dest="help", action="help",
                        help="show this help message and exit")


def parse(parser, argu):
    """
    Parse arguments given to parser

    Parameters
    ----------
    parser : argparse.ArgumentParser
        the parser used
    argu : [str]
        command-line given by user, to parse using parser

    Returns
    -------
    argparse.Namespace
        Parsed arguments
    """
    return parser.parse_args(argu)


if __name__ == '__main__':
    import argparse

    myparser = argparse.ArgumentParser(description="Align Core/Persistent families",
                                       add_help=False)

    build_parser(myparser)
    OPTIONS = parse(myparser, sys.argv[1:])
    main_from_parse(OPTIONS)



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Function to infer a phylogenetic tree from a persistent genome, using FastTree

@author gem
June 2017
"""

import os
import logging

from PanACoTA import utils

logger = logging.getLogger("tree.fasttree")

def run_tree(alignfile, boot, outdir, quiet, threads, **kwargs):
    """
    Run FastTree for the given alignment file and options

    Parameters
    ----------
    alignfile: str
        path to file containing all persistent families aligned, and grouped by genome
    boot: int or None
        number of bootstraps to calculate, None if no bootstrap asked
    outdir: str or None
        Path to the tree file that must be created
    quiet: bool
        True if nothing must be printed to stderr/stdout, False otherwise
    threads: int
        Maximum number of threads to use
    model: str
        DNA substitution model chosen by user
    kwargs: Object
        Used to be compatible with the 'run_tree' function of other softs like fastME and
        quicktree, which require more arguments
    """
    model = kwargs["model"]
    define_nb_threads(threads)
    run_fasttree(alignfile, boot, outdir, model, quiet)


def define_nb_threads(threads):
    """
    With FastTree, number of threads to use must be defined before running the
    script, by changing an environment variable.

    Parameters
    ----------
    threads: int
        Maximal number of threads to use
    """
    os.environ["OMP_NUM_THREADS"] = str(threads)


def run_fasttree(alignfile, boot, outdir, model, quiet):
    """
    Run FastTree on given alignment

    Parameters
    ----------
    alignfile: str
        Path to file containing all families aligned, grouped by genome
    boot: int or None
        Number of bootstraps to calculate (None if no bootstrap asked)
    treefile: str or None
        Path to the tree file that must be created
    model: str
        DNA substitution model
    quiet: bool
        True if nothing must be printed to stderr/stdout, False otherwise
    """
    logger.info("Running FasttreeMP...")
    if not boot:
        bootinfo = "-nosupport"
    else:
        bootinfo = "-boot {}".format(boot)
    align_name = os.path.basename(alignfile)
    logfile = os.path.join(outdir, align_name + ".fasttree.log")
    treefile = os.path.join(outdir, align_name + ".fasttree_tree.nwk")
    cmd = f"FastTreeMP -nt {model} -noml -nocat {bootinfo} -log {logfile} {alignfile}"
    logger.details("Fasttree command: " + cmd)
    if quiet:
        fnull = open(os.devnull, 'w')
    else:
        fnull = None
    stdout = open(treefile, "w")
    error = ("Problem while running Fasttree. See log file ({}) for "
             "more information.").format(logfile)
    utils.run_cmd(cmd, error, stdout=stdout, eof=True, logger=logger, stderr=fnull)



#!/usr/bin/env python3
# coding: utf-8

"""
Functional tests for the parser of tree subcommand
"""
import argparse
import pytest

from PanACoTA.subcommands import tree


def test_parser_noarg(capsys):
    """
    Test that when the script is called without any argument, an error message appears,
    indicating the required arguments.
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    with pytest.raises(SystemExit):
        tree.parse(parser, "".split())
    _, err = capsys.readouterr()
    assert "usage: " in err
    assert "-a ALIGNMENT -o OUTDIR" in err
    assert "[-s {fasttree,fastme,quicktree,iqtree,iqtree2}] [-b BOOT]" in err
    assert "[--threads THREADS] [-m MODEL]" in err
    assert "[-B] [--mem MEMORY" in err
    assert "[-v]" in err
    assert "[-q] [-h]" in err
    assert "the following arguments are required: -a, -o" in err


def test_parser_threadnotint(capsys):
    """
    Test that when the number of threads given is not an int, it returns an error message
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    with pytest.raises(SystemExit):
        tree.parse(parser, "-a align -o outdir --threads 1.2".split())
    _, err = capsys.readouterr()
    assert "argument --threads threads: invalid int value: 1.2" in err


def test_parser_thread_toomany(capsys):
    """
    Test that when the number of threads given is higher than the total number of threads,
    it returns the expected error message
    """
    import multiprocessing
    nb = multiprocessing.cpu_count()
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    with pytest.raises(SystemExit):
        tree.parse(parser, "-a align --threads {} -o outdir".format(nb + 3).split())
    _, err = capsys.readouterr()
    assert ("You have {} threads on your computer, you cannot ask for more: "
            "invalid value: {}".format(nb, nb+3)) in err


def test_parser_thread_neg(capsys):
    """
    Test that when the number of threads given is a negative number, it returns the expected
    error message.
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    with pytest.raises(SystemExit):
        tree.parse(parser, "-a align --threads -5".split())
    _, err = capsys.readouterr()
    assert ("Please provide a positive number of threads (or 0 for all threads): "
            "Invalid value: -5") in err


def test_parser_boot_notenough(capsys):
    """
    Test that when the number of threads given is a negative number, it returns the expected
    error message.
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    with pytest.raises(SystemExit):
        tree.parse(parser, "-a align -b 10 -o outdir".split())
    _, err = capsys.readouterr()
    assert ("With IQtree, number of replicates for bootstraps must be >= 1000") in err


def test_parser_quicktree_parallel(capsys):
    """
    Test that when soft is quicktree, and e ask for more than 1 thread, it returns the expected
    error message.
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    with pytest.raises(SystemExit):
        tree.parse(parser, "-a align -s quicktree --threads 2 -o outdir".split())
    _, err = capsys.readouterr()
    assert ("You cannot run quicktree with multiple threads. Choose another software, or remove "
            "the --threads option.") in err


def test_parser_quicktree_model(capsys):
    """
    Test that when soft is quicktree, and we ask for a specific model, it returns the expected
    error message.
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    with pytest.raises(SystemExit):
        tree.parse(parser, "-o outdir -a align -s quicktree -m F84".split())
    _, err = capsys.readouterr()
    assert ("Quicktree only runs the NJ algorithm. You cannot choose a DNA substitution "
            "model.") in err


def test_parser_quicktree_writeboot(capsys):
    """
    Test that when soft is quicktree, and we ask for a specific model, it returns the expected
    error message.
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    with pytest.raises(SystemExit):
        tree.parse(parser, "-o outdir -a align -s quicktree -B".split())
    _, err = capsys.readouterr()
    assert "'-B' option is only available with FastME and IQtree." in err


def test_parser_fastme_wrongmodel(capsys):
    """
    Test that when soft is fastme, and we ask for GTR model, it returns the expected
    error message.
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    with pytest.raises(SystemExit):
        tree.parse(parser, "-a align -o outdir -s fastme -m GTR".split())
    _, err = capsys.readouterr()
    assert ("GTR is not an available model for fastme. Please choose an available DNA model (see "
            "-h for more details)") in err


def test_parser_fasttree_wrongmodel(capsys):
    """
    Test that when soft is fasttree, and we ask for RY model, it returns the expected
    error message.
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    with pytest.raises(SystemExit):
        tree.parse(parser, "-a align -o outdir -s fasttree -m RY".split())
    _, err = capsys.readouterr()
    assert ("RY is not an available model for fasttree. Please choose an available DNA model (see "
            "-h for more details)") in err


def test_parser_iqtree_wrongmodel(capsys):
    """
    Test that when soft is fasttree, and we ask for RY model, it returns the expected
    error message.
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    with pytest.raises(SystemExit):
        tree.parse(parser, "-a align -o outdir -s iqtree -m toto".split())
    _, err = capsys.readouterr()
    assert ("toto is not an available model for iqtree. Please choose an available DNA model (see "
            "-h for more details)") in err


def test_parser_fasttree_writeboot(capsys):
    """
    Test that when soft is fasttree, and we ask to write bootstraps, it returns the expected
    error message.
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    with pytest.raises(SystemExit):
        tree.parse(parser, "-o outdir -a align -s fasttree -B".split())
    _, err = capsys.readouterr()
    assert "'-B' option is only available with FastME and IQtree" in err


def test_parser_fasttree_fast(capsys):
    """
    Test that when soft is fasttree, and we ask to write bootstraps, it returns the expected
    error message.
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    with pytest.raises(SystemExit):
        tree.parse(parser, "-o outdir -a align -s fasttree -fast".split())
    _, err = capsys.readouterr()
    assert "-fast option is available only for IQtree, and not compatible with '-B' and '-b' options (bootstraps)" in err


def test_parser_fastme_memory(capsys):
    """
    Test that when soft is fastme, and we ask for a memory amount, it returns the expected
    error message.
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    with pytest.raises(SystemExit):
        tree.parse(parser, "-o outdir -a align -s fastme --mem 10GB".split())
    _, err = capsys.readouterr()
    assert "'--mem' option is only available for IQtree" in err


def test_parser_iqtree_boot_fast(capsys):
    """
    Test that when soft is fasttree, and we ask to write bootstraps, it returns the expected
    error message.
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    with pytest.raises(SystemExit):
        tree.parse(parser, "-o outdir -a align -b 1000 -fast".split())
    _, err = capsys.readouterr()
    assert "-fast option is available only for IQtree, and not compatible with '-B' and '-b' options (bootstraps)" in err


def test_parser_iqtree_writeboot_fast(capsys):
    """
    Test that when soft is fasttree, and we ask to write bootstraps, it returns the expected
    error message.
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    with pytest.raises(SystemExit):
        tree.parse(parser, "-o outdir -a align -B -fast".split())
    _, err = capsys.readouterr()
    assert "-fast option is available only for IQtree, and not compatible with '-B' and '-b' options (bootstraps)" in err


def test_parser_default():
    """
    Test that when giving only align file, it returns the expected values for all other arguments
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    args = tree.parse(parser, "-a align -o outdir".split())
    assert args.alignment == "align"
    assert args.boot is None
    assert args.outdir == "outdir"
    assert args.soft == "iqtree2"
    assert args.model == "GTR"
    assert args.write_boot is False
    assert args.threads == 1
    assert args.verbose == 0
    assert args.quiet == False


def test_parser_default_fastme():
    """
    Test that when giving only align file, and to use fastme, it returns the expected values for
    all other arguments (with default DNA model)
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    args = tree.parse(parser, "-a align -o outdir -s fastme".split())
    assert args.alignment == "align"
    assert args.boot is None
    assert args.outdir == "outdir"
    assert args.soft == "fastme"
    assert args.model == "T"
    assert args.write_boot is False
    assert args.threads == 1
    assert args.verbose == 0
    assert args.quiet == False


def test_parser_default_fasttree():
    """
    Test that when giving only align file, and to use fastme, it returns the expected values for
    all other arguments (with default DNA model)
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    args = tree.parse(parser, "-a align -o outdir -s fasttree".split())
    assert args.alignment == "align"
    assert args.boot is None
    assert args.outdir == "outdir"
    assert args.soft == "fasttree"
    assert args.model == "-gtr"
    assert args.write_boot is False
    assert args.threads == 1
    assert args.verbose == 0
    assert args.quiet == False


def test_parser_all_threads():
    """
    Test that when giving alignment file, and 0 to use all threads, it returns expected values
    for all arguments
    """
    import multiprocessing
    nb = multiprocessing.cpu_count()
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    args = tree.parse(parser, "-a align -o outdir --threads 0".split())
    assert args.alignment == "align"
    assert args.boot is None
    assert args.outdir == "outdir"
    assert args.soft == "iqtree2"
    assert args.model == "GTR"
    assert args.write_boot is False
    assert args.threads == nb
    assert args.verbose == 0
    assert args.quiet == False


def test_parser_threads_ok():
    """
    Test that when giving align file and a correct number of threads (positive int), it returns the
    expected values for all arguments
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    import multiprocessing
    nb = multiprocessing.cpu_count()
    args = tree.parse(parser, f"-a align -o outdir --threads {nb}".split())
    assert args.alignment == "align"
    assert args.boot is None
    assert args.outdir == "outdir"
    assert args.soft == "iqtree2"
    assert args.model == "GTR"
    assert args.write_boot is False
    assert args.threads == nb
    assert args.verbose == 0
    assert args.quiet == False


def test_parser_fastme_modelkey():
    """
    Test that when giving align file, using fastme, with p-distance model, it returns expected
    values for all arguments
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    args = tree.parse(parser, "-a align -o outdir -s fastme -m p-distance".split())
    assert args.alignment == "align"
    assert args.boot is None
    assert args.outdir == "outdir"
    assert args.soft == "fastme"
    assert args.model == "p"
    assert args.write_boot is False
    assert args.threads == 1
    assert args.verbose == 0
    assert args.quiet == False


def test_parser_fastme_modelval():
    """
    Test that when giving align file, using fastme, with 'Y' model, it returns expected
    values for all arguments
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    args = tree.parse(parser, "-a align -o outdir -s fastme -m Y".split())
    assert args.alignment == "align"
    assert args.boot is None
    assert args.outdir == "outdir"
    assert args.soft == "fastme"
    assert args.model == "Y"
    assert args.write_boot is False
    assert args.threads == 1
    assert args.verbose == 0
    assert args.quiet == False


def test_parser_fasttree_jc():
    """
    Test that when giving align file, using fasttree, with JC model, it returns expected
    values for all arguments
    """
    parser = argparse.ArgumentParser(description="Tree", add_help=False)
    tree.build_parser(parser)
    args = tree.parse(parser, "-a align -o outdir -s fasttree -m JC".split())
    assert args.alignment == "align"
    assert args.boot is None
    assert args.outdir == "outdir"
    assert args.soft == "fasttree"
    assert args.model == ""
    assert args.write_boot is False
    assert args.threads == 1
    assert args.verbose == 0
    assert args.quiet == False



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Functions to generate the matrix pan_quali, pan_quanti, as well
as a summary file for the pangenome.

@author gem
April 2017
"""
import logging
import numpy as np

from PanACoTA import utils
from PanACoTA import utils_pangenome as utilsp

logger = logging.getLogger("pangenome.post-treat")


def post_treat(families, pangenome):
    """
    From clusters = {num: [members]}, create:

    - a pan_quali matrix (lines = families, columns = genomes, 1 if genome present in\
     family, 0 otherwise)
    - a pan_quanti matrix (lines = families, columns = genomes, number of members from given\
     genome in the given family)
    - a summary file: lines = families. For each family:

        - nb_members: total number of members
        - sum_quanti: should be the same as nb_members!
        - sum_quali: number of different genomes in family
        - nb_0: number of missing genomes in family
        - nb_mono: number of genomes with exactly 1 member
        - nb_multi: number of genomes with more than 1 member
        - sum_0-mono-multi: should be equal to the total number of genomes in dataset
        - max_multi: maximum number of members from 1 genome

    Parameters
    ----------
    families : dict
        {num_fam: [list of members]}. Can be None, and then they will be retrieved from the\
        pangenome file
    pangenome : str
        file containing pangenome
    """
    fams_by_strain, families, all_strains = utilsp.read_pangenome(pangenome, logger, families)
    open_outputs_to_write(fams_by_strain, families, all_strains, pangenome)
    # result of open_outputs_to_write = (qualis, quantis, summaries)


def open_outputs_to_write(fams_by_strain, families, all_strains, pangenome):
    """
    Open output files, and call function to generate the matrix and summary file,
    and write it in those output files

    Parameters
    ----------
    fams_by_strain : dict
        {fam_num: {strain: [members]}}
    families : dict
        {fam_num: [all members]}
    all_strains : list
        list of all genome names
    pangenome : str
        filename containing pangenome. Will be extended for the 3 output files

    Returns
    -------
    (qualis, quantis, summaries) : tuple

        with:

        - qualis = {fam_num: [0 if no gene for species, 1 if at least 1 gene, for each\
          species in all_strains]}
        - quantis = {fam_num: [number of genes for each strain in all_strains]}
        - summaries = {fam_num: [nb_members, sum_quanti, sum_quali,\
          nb_0, nb_mono, nb_multi, sum_0-mono-multi, max_multi]}

    """
    panquali = pangenome + ".quali.txt"
    panquanti = pangenome + ".quanti.txt"
    pansum = pangenome + ".summary.txt"
    with open(pansum, "w") as psf:
        psf.write("num_fam,nb_members,sum_quanti,sum_quali,"
                  "nb_0,nb_mono,nb_multi,sum_0_mono_multi,max_multi\n")
        res = generate_and_write_outputs(fams_by_strain, families,
                                         all_strains, panquali, panquanti, psf)
    return res


def generate_and_write_outputs(fams_by_strain, families, all_strains, panquali, panquanti, psf):
    """
    From the python objects of pangenome, generate qualitative and quantitative matrix,
    as well as summary file.

    Parameters
    ----------
    fams_by_strain : dict
        {fam_num: {strain: [members]}}
    families : dict
        {fam_num: [all members]}
    all_strains : list
        list of all strains
    pqlf : _io.TextIOWrapper
        open file where qualitative matrix will be written
    pqtf : _io.TextIOWrapper
        open file where quantitative matrix will be written
    psf : _io.TextIOWrapper
        open file where summary will be written

    Returns
    -------
    (qualis, quantis, summaries) : tuple

        with:

        - qualis = {fam_num: [0 if no gene for species, 1 if at least 1 gene, for each\
        species in all_strains]}
        - quantis = {fam_num: [number of genes for each strain in all_strains]}
        - summaries = {fam_num: [nb_members, sum_quanti, sum_quali,\
         nb_0, nb_mono, nb_multi, sum_0-mono-multi, max_multi]}

    """
    logger.info("Generating qualitative and quantitative matrix, and summary file")

    # Matrix has:
    # - 1 row per family (header will be added after)
    # - 1 column for fam nums + 1 column per strain
    matrix_quali = np.empty((len(families), len(all_strains) + 1), dtype=int)
    matrix_quanti = np.empty((len(families), len(all_strains) + 1), dtype=int)

    # also save matrix as python objects
    qualis = {}
    quantis = {}
    summaries = {}
    row = 0
    for fam_num in sorted(fams_by_strain, key=lambda x: int(x)):
        strains = fams_by_strain[fam_num]
        quali = [1 if strain in strains else 0 for strain in all_strains]
        quanti = [len(strains[strain]) if strain in strains else 0 for strain in all_strains]
        nb_0 = quanti.count(0)
        nb_mono = quanti.count(1)
        nb_multi = len(quanti) - nb_0 - nb_mono
        max_multi = max(quanti)
        # Add line to quali and quanti matrices
        matrix_quali[row,:] = [fam_num] + quali
        matrix_quanti[row,:] = [fam_num] + quanti
        # Write summary line
        summ = [len(families[fam_num]), sum(quanti), sum(quali),
                nb_0, nb_mono, nb_multi, nb_0 + nb_mono + nb_multi, max_multi]
        psf.write(f"{fam_num},{utils.list_to_str(summ, sep=',')}")
        # Complete python objects with quali, quanti, sumary
        qualis[fam_num] = quali
        quantis[fam_num] = quanti
        summaries[fam_num] = summ
        row += 1
    # Add headers to quali and quanti matrix
    header = np.array(["fam_num"] + all_strains)
    matrix_quali = np.vstack((header, matrix_quali))
    matrix_quanti = np.vstack((header, matrix_quanti))
    # Transpose matrix: lines = genomes, columns = families
    tmatrix_quali = matrix_quali.transpose()
    np.savetxt(panquali, tmatrix_quali, delimiter=",", fmt="%s")
    tmatrix_quanti = matrix_quanti.transpose()
    np.savetxt(panquanti, tmatrix_quanti, delimiter=",", fmt="%s")
    return qualis, quantis, summaries



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

import sys
import os
import logging
import progressbar

from PanACoTA import utils

logger = logging.getLogger("align.extract")


def get_all_seqs(all_genomes, dname, dbpath, listdir, aldir, all_fams, quiet):
    """
    For all genomes, extract its proteins present in a persistent family to the file
    corresponding to this family.

    Parameters
    ----------
    all_genomes : []
        list of all genome names
    dname : str
        name of dataset
    dbpath : str
        path to folder containing 'Proteins' and 'Genes' folders
    listdir : str
        path to folder containing the lists of proteins/genes to extract
    aldir : str
        path to folder where extracted proteins/genes must be saved
    all_fams : []
        list of all family numbers
    quiet : bool
        True if nothin must be written to stdout/stderr, False otherwise
    """
    # Get list of files not already existing
    files_todo = check_existing_extract(all_fams, aldir, dname)
    if len(files_todo) == 0:
        logger.info(("All extraction files already existing (see detailed log for "
                     "more information)"))
        logger.warning("All prt and gene files for all families already exist. The program "
                       "will use them for the next step. If you want to re-extract a given "
                       "family, remove its prt and gen extraction files. If you want to "
                       "re-extract all families, use option -F (or --force).")
        return
    logger.info("Extracting proteins and genes from all genomes")
    nbgen = len(all_genomes)
    bar = None
    curnum = 1
    if not quiet:
        widgets = ['Extraction:',
                   progressbar.Bar(marker='█', left='', right='', fill=' '),
                   ' ', progressbar.Counter(), "/{}".format(nbgen), ' (',
                   progressbar.Percentage(), ") - ", progressbar.Timer(), ' ',
                   progressbar.ETA()]
        bar = progressbar.ProgressBar(widgets=widgets, max_value=nbgen, term_width=79).start()
        curnum = 1
    for genome in all_genomes:
        ge_gen = os.path.join(listdir, dname + "-getEntry_gen_" + genome + ".txt")
        ge_prt = os.path.join(listdir, dname + "-getEntry_prt_" + genome + ".txt")
        logger.details("Extracting proteins and genes from {}".format(genome))
        prtdb = os.path.join(dbpath, "Proteins", genome + ".prt")
        gendb = os.path.join(dbpath, "Genes", genome + ".gen")
        get_genome_seqs(prtdb, ge_prt, files_todo)
        get_genome_seqs(gendb, ge_gen, files_todo)
        if not quiet:
            bar.update(curnum)
            curnum += 1
    if not quiet:
        bar.finish()


def check_existing_extract(all_fams, aldir, dname):
    """
    For each family, check if its prt and gen extraction file already exist.
    If both exist, no need to re-extract for those families.
    If only one or no one exists, put to list to extract.

    Parameters
    ----------
    all_fams : list
        list of all family numbers
    aldir : str
        path to directory where extraction files must be saved
    dname : str
        name of the dataset

    Returns
    -------
    []
        list of files that must be generated (prt and gen files)
    """
    extract_fams = []
    for fam in all_fams:
        genfile = os.path.join(aldir, "{}-current.{}.gen".format(dname, fam))
        prtfile = os.path.join(aldir, "{}-current.{}.prt".format(dname, fam))
        if not os.path.isfile(genfile) or not os.path.isfile(prtfile):
            # At least 1 file missing: re-extract all proteins and all genes
            utils.remove(genfile)
            utils.remove(prtfile)
            # As we re-extract proteins and genes, redo alignments
            mafft_file = os.path.join(aldir, "{}-mafft-align.{}.aln".format(dname, fam))
            btr_file = os.path.join(aldir, "{}-mafft-prt2nuc.{}.aln".format(dname, fam))
            utils.remove(mafft_file)
            utils.remove(btr_file)
            extract_fams.append(genfile)
            extract_fams.append(prtfile)
    # If we re-extract at least 1 family, redo the final files (concatenation and group by genome)
    if len(extract_fams) > 0:
        concat = os.path.join(aldir, "{}-complete.cat.aln".format(dname))
        outdir = os.path.split(aldir)[0]
        treedir = os.path.join(outdir, "Phylo-" + dname)
        grpfile = os.path.join(treedir, dname + ".grp.aln")
        utils.remove(concat)
        utils.remove(grpfile)
    return extract_fams


def get_genome_seqs(fasta, tabfile, files_todo, outfile=None):
    """
    From a fasta file, extract all sequences given in the tab file.
    The tab file can contain:

    - 1 sequence name per line -> all sequences will be extracted to the same file
    - 1 sequence + 1 filename per line -> each sequence will be extracted in the given file

    If outfile not given, the tab file must contain 2 columns (1 for the sequence name,
    1 for its output file). If an outfile is given, only the 1st column of tab file
    will be considered, and all sequences will be extracted to the given outfile.

    Parameters
    ----------
    fasta : str
        path to fasta file from which sequences must be extracted
    tabfile : str
        path to the tab file containing the names of sequences to extract
    files_todo : list
        list of files which must be generated (prt and gen files). Others
        already exist, so ignore them.
    outfile : str or None
        if None, the tab file must contain 2 columns (1 for the sequence name,
        1 for its output file). If an outfile is given (not None), only the 1st column of tab file
        will be considered, and all sequences will be extracted to the given outfile.
    """
    with open(tabfile, "r") as tabf:
        to_extract = get_names_to_extract(tabf, outfile)
    if outfile:
        if os.path.isfile(outfile):
            logger.warning("Sequences are already extracted in {}. This will "
                           "be used for next step. If you want "
                           "to re-extract all sequences, use option -F (or "
                           "--force)".format(outfile))
            return
        with open(fasta, "r") as fasf, open(outfile, "a") as outf:
            extract_sequences(to_extract, fasf, outf=outf)
    else:
        with open(fasta, "r") as fasf:
            extract_sequences(to_extract, fasf, files_todo=files_todo)


def get_names_to_extract(tabf, outfile):
    """
    From the tab file, get names of sequences to extract.

    Parameters
    ----------
    tabf : _io.TextIO
        open file containing names of sequences to extract
    outfile : str or None
        if None, the tab file must contain 2 columns (1 for the sequence name,
        1 for its output file). If an outfile is given (not None), only the 1st column of tab file
        will be considered, and all sequences will be given 'outfile' as output file

    Returns
    -------
    dict
        {sequence_to_extract: file_to_which_it_will_be_extracted}
    """
    to_extract = {}
    for line in tabf:
        if outfile:
            seq = line.split()[0].strip()
            out = outfile
        else:
            try:
                seq, out = line.split()[:2]
            except ValueError:
                logger.error(("Your file {} does not contain an output filename for {}. "
                              "Please give an output filename for each sequence to extract, "
                              "or give a general output filename where all sequences will "
                              "be extracted.").format(tabf.name, line.strip()))
                sys.exit(1)
        to_extract[seq] = out
    return to_extract


def extract_sequences(to_extract, fasf, files_todo=None, outf=None):
    """
    Extract sequences from an open fasta file 'fasf', and a list of sequences to
    extract

    Parameters
    ----------
    to_extract : dict or []
        {sequence_to_extract: file_to_which_it_will_be_extracted} or list of sequences to
        extract, all in a same outfile (name must be given in 'outf')
    fasf : _io.TextIO
        open file containing sequences in multi-fasta format
    files_todo : list or None
        list of files which must be generated (prt and gen files). Others
        already exist, so ignore them.
    outf : _io.TextIO or None
        If an outfile is given (not None), and 'to_extract' is a dict, only its keys will be
        considered, and all these sequences will be extracted to 'outfile' (if 'to_extract' is a
        list, will extract all sequences of this list). Otherwise, if None,
        each sequence will be extracted to its corresponding value in 'to_extract'.
    """

    # Create optimized index for requests
    if files_todo is None:
        files_todo = []
    files_todo = frozenset(files_todo)
    if type(to_extract) == list:
        to_extract = frozenset(to_extract)

    # State machine variables
    previous_fp = None

    for line in fasf:
        if line[0] == '>':
            # Close previous file if needed
            if outf is None and previous_fp is not None:
                previous_fp.close()
            previous_fp = None

            # Extract sequence name
            last_char = line.find(' ')
            if last_char == -1:
                last_char = len(line)
            seq = line[1:last_char].strip()

            # Seq is part of sequences to extract
            if seq in to_extract:
                # Open the right file
                previous_fp = outf
                if previous_fp is None:
                    out = to_extract[seq]
                    if out in files_todo:
                        previous_fp = open(out, "a")
                    else:
                        print(f"Sequence {seq} not written because no output file specified", file=sys.stderr)

        # Write the line content if the file pointer is opened
        if previous_fp is not None:
            previous_fp.write(line)

    if outf is None and previous_fp is not None:
        previous_fp.close()
